/*
 * Copyright (C) 2014 greilich
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package fntd.io;

import fntd.stacks.StackOperations;

import ij.IJ;
import ij.ImagePlus;
import ij.ImageStack;
import ij.gui.GenericDialog;
import ij.io.DirectoryChooser;
import ij.io.Opener;
import ij.measure.ResultsTable;
import ij.plugin.HyperStackConverter;
import ij.plugin.TextReader;
import ij.process.ImageProcessor;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import static java.lang.Math.floor;
import java.util.Arrays;

/**
 * Class to import data from the FXR700RG reader Usage: FXR700Importer.run() As
 * plugin FXR700Importer.runSilent() As plugin
 *
 * Stores the images according to user settings into a hyperstack with
 * dimensions: tiles scanned frames in x,y on a detector slices scanned frames
 * in z on a detector channels fluorescence (APD) and reflected (PD) image frame
 * detectors * @author greilich
 */
public class FXR700Importer {

    private enum DataType {

        PNG, TXT
    }

    private static String dir = "";
    private static int nX = 0;
    private static int nY = 0;
    private static int nZ = 0;
    private static int nDetectors = 0;
    private static boolean flOnly = true;
    private static boolean showList = false;
    private static boolean stitchTiles = false;
    private static boolean asStack = false;
    private static boolean silent = false;
    private static DataType dataType = DataType.TXT;
    private static ResultsTable rt = null;

    public static ImagePlus[] run() {
        if (getDir() != 0) {
            return (null);
        }

        int result = readConfigFile();
        if (result == -1) {
            IJ.showMessage("File Configuration.ini not found - please set values manually.");
        }
        if (result == -2) {
            IJ.showMessage("Error reading file Configuration.ini not found - please set values manually.");
        }

        result = getOptions();
        if (result != 0) {
            return (null);
        }

        rt = new ResultsTable();
        ImagePlus[] resImg = importFXR700Data();

        if (showList) {
            rt.show("File list");
        }

        return (resImg);
    }

    public static ImagePlus[] runSilent(String directory) {
        silent = true;
        showList = false;
        asStack = false;
        dir = directory;

        int result = readConfigFile();
        if (result != 0) {
            return (null);
        }

        return (importFXR700Data());
    }

    private static int getDir() {
        DirectoryChooser dc = new DirectoryChooser("Select directory with FXR700 data");
        dir = dc.getDirectory();
        if (dir == null) {
            return -1;
        } else {
            return 0;
        }
    }

    private static int readConfigFile() {

        try {
            FileReader fr = new FileReader(dir + File.separator + "Configuration.ini");
            BufferedReader cfReader = new BufferedReader(fr);
            String aLine;
            String[] parts;
            int DX = 0;
            int DY = 0;
            while ((aLine = cfReader.readLine()) != null) {
                if (aLine.startsWith("det in x")) {
                    parts = aLine.split("=");
                    DX = Integer.parseInt(parts[1].replace("\"", "").replace(" ", ""));
                }
                if (aLine.startsWith("det in y")) {
                    parts = aLine.split("=");
                    DY = Integer.parseInt(parts[1].replace("\"", "").replace(" ", ""));
                }
                if (aLine.startsWith("x fields")) {
                    parts = aLine.split("=");
                    nX = Integer.parseInt(parts[1].replace("\"", "").replace(" ", ""));
                }
                if (aLine.startsWith("y fields")) {
                    parts = aLine.split("=");
                    nY = Integer.parseInt(parts[1].replace("\"", "").replace(" ", ""));
                }
                if (aLine.startsWith("z fields")) {
                    parts = aLine.split("=");
                    nZ = Integer.parseInt(parts[1].replace("\"", "").replace(" ", ""));
                }
            }
            nDetectors = DX * DY;
        } catch (FileNotFoundException ex) {
            return -1;
        } catch (IOException ex) {
            return -2;
        }

        return 0;
    }

    private static int getOptions() {

        String[] choices = {"Text files, 32bit float", "PNG images, 8/16bit integer"};
        String choice;
        GenericDialog gd = new GenericDialog("Import FXR700 data");
        gd.addNumericField("Number of tiles in x:  ", nX, 0);
        gd.addNumericField("Number of tiles in y:  ", nY, 0);
        gd.addNumericField("Number of slices in z: ", nZ, 0);
        gd.addNumericField("Number of detectors:   ", nDetectors, 0);
        gd.addChoice("Data type", choices, choices[0]);
        gd.addCheckbox("Read fluorescence channel only", flOnly);
        gd.addCheckbox("Stitch tile (and use frame dimension for detectors)", stitchTiles);
        gd.addCheckbox("Show file list", showList);
        gd.addCheckbox("Return images as stack rather than hyperstack", asStack);
        gd.showDialog();
        if (gd.wasCanceled()) {
            return -1;
        }

        nX = (int) gd.getNextNumber();
        nY = (int) gd.getNextNumber();
        nZ = (int) gd.getNextNumber();
        nDetectors = (int) gd.getNextNumber();
        choice = gd.getNextChoice();
        dataType = (choice.equalsIgnoreCase(choices[0])) ? DataType.TXT : DataType.PNG;
        flOnly = gd.getNextBoolean();
        stitchTiles = gd.getNextBoolean();
        showList = gd.getNextBoolean();
        asStack = gd.getNextBoolean();

        return 0;
    }

    private static ImagePlus[] importFXR700Data() {

        File folder = new File(dir);
        File[] listOfFiles = folder.listFiles();
        Arrays.sort(listOfFiles);
        int nChannels = flOnly ? 1 : 2;

        int imageCounterFl = 0;
        int[] imageIndicesFl = new int[listOfFiles.length];
        int imageCounterRe = 0;
        int[] imageIndicesRe = new int[listOfFiles.length];
        for (int i = 0; i < listOfFiles.length; i++) {
            if (listOfFiles[i].isFile()) {
                String filename = listOfFiles[i].getName();
                String extension = filename.substring(filename.lastIndexOf(".") + 1, filename.length());
                if (extension.equalsIgnoreCase(dataType.toString())) {
                    rt.incrementCounter();
                    rt.addValue("List index", i);
                    rt.addValue("File name", filename);
                    String date = filename.substring(0, 4);
                    rt.addValue("Date", date);
                    String time = filename.substring(5, 14);
                    rt.addValue("Time", time);
                    String channel = filename.substring(14, 16);
                    rt.addValue("Channel", channel);
                    if (channel.equalsIgnoreCase("fl")) {
                        imageIndicesFl[imageCounterFl++] = i;
                    }
                    if (channel.equalsIgnoreCase("re")) {
                        imageIndicesRe[imageCounterRe++] = i;
                    }
                }
            }
        }

        /* Check validity */
        String msg;
        if ((!flOnly) && (imageCounterFl != imageCounterRe)) {
            msg = "Different number of images in channels. Aborting...";
            if (!silent) {
                IJ.showMessage(msg);
            } else {
                System.out.println(msg);
            }
            return (null);
        }

        if (imageCounterFl > nDetectors * nZ * nY * nX) {
            msg = "Found more images than expected from given dimensions. Will cut list of images to match.";
            if (!silent) {
                IJ.showMessage(msg);
            } else {
                System.out.println(msg);
            }
            imageCounterFl = nDetectors * nZ * nY * nX;
            if (!flOnly) {
                imageCounterRe = nDetectors * nZ * nY * nX;
            }
        }

        if (imageCounterFl < nDetectors * nZ * nY * nX) {
            msg = "Found fewer images than expected from given dimensions. Will cut number of detectors to match.";
            if (!silent) {
                IJ.showMessage(msg);
            } else {
                System.out.println(msg);
            }
            nDetectors = (int) floor((double) imageCounterFl / (double) (nZ * nY * nX));
            imageCounterFl = nDetectors * nZ * nY * nX;
            if (!flOnly) {
                imageCounterRe = nDetectors * nZ * nY * nX;
            }
        }

        ImagePlus[] tmpImg = new ImagePlus[1];
        ImageProcessor tmpIP;

        TextReader textReader = null;
        Opener opener = null;
        if (dataType == DataType.TXT) {
            textReader = new TextReader();
            tmpImg[0] = new ImagePlus("Import", textReader.open(listOfFiles[imageIndicesFl[0]].getAbsolutePath()));
        } else {
            opener = new Opener();
            tmpImg[0] = opener.openImage(listOfFiles[imageIndicesFl[0]].getAbsolutePath());
        }


        ImageStack tmpIms = new ImageStack(tmpImg[0].getWidth(), tmpImg[0].getHeight());
        for (int i = 0; i < imageCounterFl; i++) {
            if (!silent) {
                IJ.showProgress(i, imageCounterFl - 1);
                IJ.showStatus("Reading in all fluorescence images (" + i + " of " + imageCounterFl + ").");
            }else{
                    if((i/imageCounterRe)%10 == 0){
                        System.out.print("Read "+ (int)(i/imageCounterRe*100) + "% of fluorescence images.");
                    }
            }
            if (dataType == DataType.TXT) {
                tmpIP = textReader.open(listOfFiles[imageIndicesFl[i]].getAbsolutePath());
            } else {
                tmpIP = opener.openImage(listOfFiles[imageIndicesFl[i]].getAbsolutePath()).getProcessor();
            }
            
            tmpIP.flipVertical();
            tmpIms.addSlice(listOfFiles[imageIndicesFl[i]].getName(), tmpIP);
        }
        if (!flOnly) {
            for (int i = 0; i < imageCounterRe; i++) {
                if (!silent) {
                    IJ.showProgress(i, imageCounterRe - 1);
                    IJ.showStatus("Reading in all relected light images (" + i + " of " + imageCounterRe + ").");
                }else{
                    if((i/imageCounterRe)%10 == 0){
                        System.out.print("Read "+ (int)(i/imageCounterRe*100) + "% of relected light images.");
                    }
                }
                if (dataType == DataType.TXT) {
                    tmpIP = textReader.open(listOfFiles[imageIndicesRe[i]].getAbsolutePath());
                } else {
                    tmpIP = opener.openImage(listOfFiles[imageIndicesRe[i]].getAbsolutePath()).getProcessor();
                }

                tmpIP.flipVertical();
                tmpIms.addSlice(listOfFiles[imageIndicesRe[i]].getName(), tmpIP);
            }
        }

        tmpImg[0] = new ImagePlus("Stack import of " + dir, tmpIms);
        if (asStack||!stitchTiles) {
            if(asStack){
                return (tmpImg);
            }else{
                ImagePlus[] resImp = new ImagePlus[nDetectors];
                int curSlice = 1;
                int nSlicesPerDetector = nZ * nX * nY * nChannels;
                
                for(int i = 0; i < nDetectors; i++){
                    ImageStack resIms = new ImageStack(tmpImg[0].getWidth(), tmpImg[0].getHeight());
                    for(int j = 0; j < nSlicesPerDetector; j++){
                        tmpImg[0].setSlice(curSlice);
                        resIms.addSlice(tmpImg[0].getProcessor());
                        curSlice++;
                    }
                    resImp[i] = HyperStackConverter.toHyperStack(new ImagePlus("FXR700RG import, detector no. "+(i+1), resIms), 
                            nChannels, 
                            nZ, 
                            nX*nY, 
                            "xyztc", 
                            "grayscale");
                }
                return(resImp);
            }
        } else {
            ImagePlus[] resImg = new ImagePlus[1];
            resImg[0] = StackOperations.tileZXYTC(nX, nY, nZ, nDetectors, nChannels, tmpImg[0]);
            resImg[0].setTitle("Hyperstack import of " + dir);
            resImg[0] = HyperStackConverter.toHyperStack(resImg[0], nChannels, nZ, nDetectors, "xyztc", "grayscale");
            return(resImg);
        }
    }

}
