/*
 * Copyright (C) 2015 Jasper J.M. Kouwenberg
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package fntd.plugins;

import fntd.FNTDexception;
import fntd.tracking.FNTDtracks;
import fntd.access.DisplayImage;
import ij.IJ;
import ij.ImagePlus;
import ij.gui.NonBlockingGenericDialog;
import ij.plugin.PlugIn;
import ij.Prefs;
import ij.gui.GenericDialog;
import java.awt.Button;
import java.awt.Font;
import java.awt.Panel;
import java.awt.event.ActionEvent;

/**
 * Plugin for the feature point linker and trajectory relinker
 *
 * @author Jasper J.M. Kouwenberg
 */
public class ParticleTracker implements PlugIn {
    
    NonBlockingGenericDialog gd;
    FNTDtracks tracks;
    ImagePlus imp;
    Panel about_panel;
    boolean first_run = true;
        
    // Feature Point Detector Dialog Tab
    boolean absolute            = false;
    int rad                     = 3;
    double cut                  = 0;
    double threshold            = 1;
    double per                  = 0;

    // Feature Point Linker Dialog Tab
    int linkrange               = 2;
    int displacement            = 20;
    double drift                = 3;
    double WeightDistance       = 1;
    double WeightIntensity      = 1;
    double WeightVelocity       = 1;
    double WeightAngle          = 1;
    double max_angle            = 20;
    int angle_history           = 3;
    boolean relink              = true;
    double relink_cost          = 1;
    double relink_distance      = 40;
    
    // Filter Dialog Tab
    int min_track_length        = 4;

    @Override
    public void run(String arg) {
        // Get current image
        this.imp = IJ.getImage();
        
        while (true) {
            this.gd = new NonBlockingGenericDialog("Particle Tracker");
            this.gd.setLocation(100, 100);
                
            // Load settings
            this.absolute            = Prefs.getBoolean(".FNTD.absolute", this.absolute);
            this.rad                 = Prefs.getInt(".FNTD.rad", this.rad);
            this.cut                 = Prefs.getDouble(".FNTD.cut", this.cut);
            this.threshold           = Prefs.getDouble(".FNTD.threshold", this.threshold);

            this.linkrange           = Prefs.getInt(".FNTD.linkrange", this.linkrange);
            this.displacement        = Prefs.getInt(".FNTD.displacement", this.displacement);
            this.drift               = Prefs.getDouble(".FNTD.drift", this.drift);
            this.WeightDistance      = Prefs.getDouble(".FNTD.WeightDistance", this.WeightDistance);
            this.WeightIntensity     = Prefs.getDouble(".FNTD.WeightIntensity", this.WeightIntensity);
            this.WeightVelocity      = Prefs.getDouble(".FNTD.WeightVelocity", this.WeightVelocity);
            this.WeightAngle         = Prefs.getDouble(".FNTD.WeightAngle", this.WeightAngle);
            this.max_angle           = Prefs.getDouble(".FNTD.max_angle", this.max_angle);
            this.angle_history       = Prefs.getInt(".FNTD.angle_history", this.angle_history);
            this.relink              = Prefs.getBoolean(".FNTD.relink", this.relink);
            this.relink_cost         = Prefs.getDouble(".FNTD.relink_cost", this.relink_cost);
            this.relink_distance     = Prefs.getDouble(".FNTD.relink_distance", this.relink_distance);

            this.min_track_length    = Prefs.getInt(".FNTD.track_length", this.min_track_length);

            this.gd.addMessage       ("MOSAIC Feature Point Detection parameters", new Font("SansSerif", Font.BOLD, 15));
            this.gd.addNumericField  ("Particle radius", this.rad, 0);
            this.gd.addNumericField  ("Cut-off", this.cut, 2);
            this.gd.setInsets(-3,40,0);
            this.gd.addCheckbox      ("Absolute threshold", this.absolute);
            this.gd.addNumericField  ("Threshold*", this.threshold, 2);
            this.gd.setInsets(0,30,0);
            this.gd.addMessage       ("*Given as percentile of max. pixels \n when absolute threshold is unchecked");

            this.gd.addMessage       ("FNTD Feature Point Linking parameters", new Font("SansSerif", Font.BOLD, 15));
            this.gd.addNumericField  ("Link range", this.linkrange, 0);
            this.gd.addNumericField  ("Displacement", this.displacement, 0);
            this.gd.addNumericField  ("Drift", this.drift, 0);
            this.gd.addNumericField  ("Distance weight", this.WeightDistance, 2);
            this.gd.addNumericField  ("Intensity weight", this.WeightIntensity, 2);
            this.gd.addNumericField  ("Velocity weight", this.WeightVelocity, 2);
            this.gd.addNumericField  ("Angle weight", this.WeightAngle, 2);
            this.gd.addNumericField  ("Max. angle", this.max_angle, 0);
            this.gd.addNumericField  ("Angle history", this.angle_history, 0);
            this.gd.setInsets(-3,40,0);
            this.gd.addCheckbox      ("Trajectory relinking", this.relink);
            this.gd.addNumericField  ("Relink cost factor", this.relink_cost, 2);
            this.gd.addNumericField  ("Relink displacement", this.relink_distance, 0);
            
            this.gd.addMessage       ("FNTD track length filter", new Font("SansSerif", Font.BOLD, 15));
            this.gd.addNumericField  ("Min. track length", this.min_track_length, 0);
            
            Panel p = new Panel();
            Button about = new Button("Terms of use");
            p.add(about);

            gd.addPanel(p);

            about.addActionListener((ActionEvent arg0) -> {
                ParticleTracker.TermsOfUseDialog();
            });
            
            this.gd.showDialog();
            
            if (this.gd.wasCanceled()) return;
                
            // Retrieve input
            this.absolute        = gd.getNextBoolean();
            this.rad             = (int) gd.getNextNumber();
            this.cut             = gd.getNextNumber();
            this.threshold       = gd.getNextNumber();
            this.linkrange       = (int) gd.getNextNumber();
            this.displacement    = (int) gd.getNextNumber();
            this.drift           = gd.getNextNumber();
            this.WeightDistance  = gd.getNextNumber();
            this.WeightIntensity = gd.getNextNumber();
            this.WeightVelocity  = gd.getNextNumber();
            this.WeightAngle     = gd.getNextNumber();
            this.max_angle       = gd.getNextNumber();
            this.angle_history   = (int) gd.getNextNumber();
            this.relink          = gd.getNextBoolean();
            this.relink_cost     = gd.getNextNumber();
            this.relink_distance = gd.getNextNumber();
            this.min_track_length= (int) gd.getNextNumber();

            boolean changed = false;
            if (this.absolute != Prefs.getBoolean(".FNTD.absolute", this.absolute) || this.rad != Prefs.getInt(".FNTD.rad", this.rad)
                    || this.cut != Prefs.getDouble(".FNTD.cut", this.cut) || this.threshold != Prefs.getDouble(".FNTD.threshold", this.threshold)) {
                changed = true;
            }

            // Save settings
            Prefs.set("FNTD.absolute", this.absolute);
            Prefs.set("FNTD.rad", this.rad);
            Prefs.set("FNTD.cut", this.cut);
            Prefs.set("FNTD.threshold", this.threshold);
            Prefs.set("FNTD.linkrange", this.linkrange);
            Prefs.set("FNTD.displacement", this.displacement);
            Prefs.set("FNTD.drift", this.drift);
            Prefs.set("FNTD.WeightDistance", this.WeightDistance);
            Prefs.set("FNTD.WeightIntensity", this.WeightIntensity);
            Prefs.set("FNTD.WeightVelocity", this.WeightVelocity);
            Prefs.set("FNTD.WeightAngle", this.WeightAngle);
            Prefs.set("FNTD.max_angle", this.max_angle);
            Prefs.set("FNTD.angle_history", this.angle_history);
            Prefs.set("FNTD.relink", this.relink);
            Prefs.set("FNTD.relink_cost", this.relink_cost);
            Prefs.set("FNTD.relink_distance", this.relink_distance);
            Prefs.set("FNTD.track_length", this.min_track_length);
            Prefs.savePreferences();
            
            gd.dispose();

            if (!this.absolute) {
                this.per = this.threshold;
                this.threshold = 0;
            }

            try {
                if (first_run) {
                    this.tracks = new FNTDtracks(imp, this.rad, this.cut, this.per, this.threshold);
                    first_run = false;
                } else if (changed) {
                    this.tracks = new FNTDtracks(imp, this.rad, this.cut, this.per, this.threshold);
                } else
                    this.tracks.resetTracks();

                this.tracks.linkParticles(this.linkrange, this.displacement, this.drift, this.WeightDistance, this.WeightIntensity,
                        this.WeightVelocity, this.WeightAngle, this.max_angle, this.angle_history, this.relink, this.relink_cost, this.relink_distance);
                this.tracks.filterTrackLength(this.min_track_length);
                DisplayImage.show(imp, this.tracks);
            } catch (FNTDexception ex) { }
        }
    }
    
    public static void TermsOfUseDialog() {
        GenericDialog tou = new GenericDialog("Terms of Use");
        tou.hideCancelButton();
        
        tou.addMessage("Copyright 2015 Steffen Greillich, Jasper J.M. Kouwenberg\n"
                + "This software is released under GNU General Public License and is free to use.\n\n"
                + "Requires citation of the following papers for scientific publications:\n"
                + " - Sbalzarini, I.F. & Koumoutsakos, P., 2005. Feature point tracking\n   and trajectory analysis for 391 video imaging in cell biology.\n"
                + " - Kouwenberg, J. J. M. et al, 2015, 3D Feature Point Tracking Method for Ion Radiation");
        
        tou.showDialog();
    }
}
