/*
 * Copyright (C) 2014 Steffen Greilich
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package fntd.plugins;

/**
 *
 * @author greilich
 */
import ij.*;
import ij.gui.*;
import ij.plugin.*;
import ij.measure.ResultsTable;

import java.util.ArrayList;

public class ProcessFlatCube implements PlugIn {

    static int nX = 5;
    static int nY = 5;
    static boolean returnZAxisProfiles = true;
    static boolean returnTileWiseMeans = true;
    static boolean showPlots = false;
    static boolean returnNormalizedCubes = false;
    static boolean returnFrames = false;

    @Override
    public void run(String arg) {

        ResultsTable zAxisProfileTable = new ResultsTable();
        ResultsTable frameWiseMeansTable = new ResultsTable();

        ZProjector zProjector;
        ImageCalculator imageCalculator = new ImageCalculator();

        ArrayList<ImagePlus> frameList;
        ArrayList<ImagePlus> impList;

        ImagePlus tmpImp;
        ImagePlus resImp;
        ImagePlus resNormImp;

        ImageStack tmpStack;
        ImageStack resStack;

        /**
         * Get user input
         */
        GenericDialog gd = new GenericDialog("FlatCubeCondenser");
        gd.addNumericField("Number of tiles in x: ", nX, 0);
        gd.addNumericField("Number of tiles in y: ", nY, 0);
        gd.addChoice("Operation", ZProjector.METHODS, "Median");
        gd.addCheckbox("Return z axis profiles for resulting flat cubes", true);
        gd.addCheckbox("Return tile-wise means", true);
        gd.addCheckbox("Show plots (may look funny, only if z-axis/tile-wise are checked)", false);
        gd.addCheckbox("Return additionally flat-cubes to normalized to last image", false);
        gd.addCheckbox("Return additionally raw stacks for all frames", false);
        gd.showDialog();

        if (gd.wasCanceled()) {
            return;
        }

        nX = (int) gd.getNextNumber();
        nY = (int) gd.getNextNumber();

        String method = gd.getNextChoice();
        int methodIndex = ZProjector.MEDIAN_METHOD;
        for (int i = 0; i < ZProjector.METHODS.length; i++) {
            if (ZProjector.METHODS[i].equals(method)) {
                methodIndex = i;
            }
        }

        returnZAxisProfiles = gd.getNextBoolean();
        returnTileWiseMeans = gd.getNextBoolean();
        showPlots = gd.getNextBoolean();
        returnNormalizedCubes = gd.getNextBoolean();
        returnFrames = gd.getNextBoolean();

        /**
         * Divide original image into frames
         */
        // Get currently active image, compute width and height of a single tile
        ImagePlus superImp = IJ.getImage();
        int dX = superImp.getWidth() / nX;
        int dY = superImp.getHeight() / nY;

        frameList = new ArrayList<>();

        int nFrames = superImp.getNFrames();
        int nImagesPerFrame = superImp.getStackSize() / nFrames;

        ImageStack superStack = superImp.getImageStack();

        for (int i = 0; i < nFrames; i++) {
            tmpStack = new ImageStack(superImp.getWidth(), superImp.getHeight());

            for (int j = 0; j < nImagesPerFrame; j++) {
                tmpStack.addSlice(superStack.getProcessor(i * nImagesPerFrame + j + 1));
            }

            tmpImp = new ImagePlus("Frame " + Integer.toString(i + 1), tmpStack);
            frameList.add(tmpImp);

            if (returnFrames) {
                frameList.get(i).show();
            }
        }

        /**
         * Process all frames
         */
        for (int i = 0; i < nFrames; i++) {

            /**
             * Go through current frame and cut out the z-stacks for all tiles;
             * Collect them in ArrayList impList. At the same time, get
             * information on mean intensities
             */
            impList = new ArrayList<>();

            for (int iX = 0; iX < nX; iX++) {
                for (int iY = 0; iY < nY; iY++) {
                    frameList.get(i).setRoi(iX * dX, iY * dY, dX, dY);

                    tmpImp = new Duplicator().run(frameList.get(i));
                    impList.add(tmpImp);

                    // Compute mean intensity and add to table
                    for (int iZ = 0; iZ < tmpImp.getNSlices(); iZ++) {
                        tmpImp.setSlice(iZ + 1);
                        frameWiseMeansTable.incrementCounter();
                        frameWiseMeansTable.addValue("flat.cube.no", i + 1);
                        frameWiseMeansTable.addValue("tile.index.x", iX + 1);
                        frameWiseMeansTable.addValue("tile.index.y", iY + 1);
                        frameWiseMeansTable.addValue("tile.index.z", iZ + 1);
                        frameWiseMeansTable.addValue("mean intensity", tmpImp.getProcessor().getStatistics().mean);
                    }
                }
            }

            /**
             * Rearrange all tiles of each z-slice (for the current frame) to an
             * image stack (tmpStack) and apply z-projection to this stack,
             * yielding the averaged (median,...) image for all tiles in one
             * slice, and stack those average (...) images again to a z-stack,
             * where z-slices correspond to those of the original data.
             */
            resStack = new ImageStack(dX, dY);

            for (int j = 0; j < nImagesPerFrame; j++) {
                tmpStack = new ImageStack(dX, dY);

                for (ImagePlus impList1 : impList) {
                    impList1.setSlice(j + 1);
                    tmpStack.addSlice(impList1.getProcessor());
                }

                tmpImp = new ImagePlus("tmpImg", tmpStack);
                zProjector = new ZProjector(tmpImp);
                zProjector.setMethod(methodIndex);
                zProjector.doProjection();
                resStack.addSlice(zProjector.getProjection().getProcessor());
            }

            /**
             * Convert resulting stack to an image and show.
             */
            resImp = new ImagePlus("FlatCube " + Integer.toString(i + 1), resStack);
            resImp.setSlice(resImp.getNSlices() / 2);
            resImp.show();

            /**
             * Do z-axis profile for current frame (i.e. flat cube) and add to
             * table
             */
            for (int j = 0; j < resImp.getNSlices(); j++) {
                resImp.setSlice(j + 1);
                zAxisProfileTable.incrementCounter();
                zAxisProfileTable.addValue("flat.cube.no", i + 1);
                zAxisProfileTable.addValue("slice.no", j + 1);
                zAxisProfileTable.addValue("intensity", resImp.getProcessor().getStatistics().mean);
            }

            /**
             * Normalize current frame (i.e. flat cube) and show, if requested
             */
            if (returnNormalizedCubes) {
                resImp = new ImagePlus("FlatCube " + Integer.toString(i + 1) + " (normalized)", resStack);
                resImp.setSlice(resImp.getNSlices());
                resNormImp = new ImagePlus("Last slice", resImp.getProcessor());
                resImp = imageCalculator.run("Divide create 32-bit stack", resImp, resNormImp);
                resImp.show();
            }
        }

        /**
         * Show z axis profile result table if requested
         */
        if (returnZAxisProfiles) {
            zAxisProfileTable.show("Z axis intensity profiles");
            if (showPlots) {
                Plot zAxisProfilePlot = new Plot("Z axis intensity profiles",
                        "slice no.",
                        "intensity",
                        zAxisProfileTable.getColumn(1),
                        zAxisProfileTable.getColumn(2));
                zAxisProfilePlot.show();
            }
        }

        /**
         * Show tile-wise mean intensity table if requested
         */
        if (returnTileWiseMeans) {
            frameWiseMeansTable.show("Tile-wise mean intensity");
            zAxisProfileTable.show("Z axis intensity profiles");

            if (showPlots) {
                float[] iX = frameWiseMeansTable.getColumn(1);
                float[] iY = frameWiseMeansTable.getColumn(2);
                float[] iZ = frameWiseMeansTable.getColumn(3);
                for (int i = 0; i < iX.length; i++) {
                    iX[i] = iX[i] + (iY[i] - 1.0f) * nY + (iZ[i] - 1.0f) * nX * nY;
                }

                Plot frameWiseMeansPlot = new Plot("Tile-wise mean intensity",
                        "tile no.",
                        "intensity",
                        iX,
                        frameWiseMeansTable.getColumn(4));
                frameWiseMeansPlot.show();
            }
        }

    }
}
