/*
 * Copyright (C) 2014 greilich
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU StackOperations Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU StackOperations Public License for more details.
 *
 * You should have received a copy of the GNU StackOperations Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package fntd.stacks;

import fntd.io.SaveImage;
import ij.IJ;
import ij.ImagePlus;
import ij.ImageStack;
import ij.plugin.Duplicator;
import ij.plugin.HyperStackConverter;
import ij.process.FloatProcessor;
import ij.process.ImageConverter;
import ij.process.ImageProcessor;
import java.io.File;

/**
 *
 * @author greilich
 */
public class StackOperations {

    /**
     * Transforms image stack into a single image or new stack by tiling
     * successive slices to new images, e.g. transform a 100-image stack into 25
     * slices with 2x2 tiles. The input image stack (as ImagePlus) is assumed to
     * have only one channel and one frame (i.e. no hyperstack), instead it has
     * been serialized using the order ZXYTC.
     *
     * @param nX Number of tiles in x
     * @param nY Number of tiles in y
     * @param nZ Number of slices in z
     * @param nT Number of frames
     * @param nC Number of channels
     * @param imp The image stack to be processed
     * @return The new stack
     */
    static public ImagePlus tileZXYTC(int nX, int nY, int nZ, int nT, int nC, ImagePlus imp) {
        IJ.showStatus("Tiling stack.");

        int dX = imp.getWidth();
        int dY = imp.getHeight();
        int nSlices = imp.getNSlices();

        if (nX * nY * nZ * nT * nC != nSlices) {
            return null;
        }

        ImageStack resIms = new ImageStack(nX * dX, nY * dY);
        ImagePlus tmpImp;

        // TODO: fully implement zero image padding
        int baseIndex;
        for (int iC = 0; iC < nC; iC++) {
            for (int iT = 0; iT < nT; iT++) {
                for (int iZ = 0; iZ < nZ; iZ++) {

                    // first index of current image
                    baseIndex = iC * nT * nZ * nY * nX
                            + iT * nZ * nY * nX
                            + iZ + 1;

                    tmpImp = new ImagePlus("tmp", new FloatProcessor(nX * dX, nY * dY));
                    IJ.showProgress(baseIndex, nSlices);

                    for (int iY = 0; iY < nY; iY++) {
                        for (int iX = 0; iX < nX; iX++) {

                            imp.setSlice(baseIndex + nZ * (nX * iY + iX));
                            IJ.run(imp, "Copy", "");

                            tmpImp.setRoi(iX * dX, iY * dY, dX, dY);
                            IJ.run(tmpImp, "Paste", "");
                        }
                    }
                    resIms.addSlice(tmpImp.getProcessor());
                }
            }
        }
        IJ.showStatus("Done.");
        return (new ImagePlus(imp.getTitle() + " - tiled", resIms));
    }

    /**
     * Transforms image stack into a single image or new stack by tiling
     * successive slices to new images, e.g. transform a 100-image stack into 25
     * slices with 2x2 tiles. If the stack has additional dimensions, i.e.
     * multiple frames or channels, they will be maintained in the process. If
     * the number of slices is not a multiple of nX*nY, the last tiles will be
     * filled with zero values.
     *
     * @param nX Number of tiles in x
     * @param nY Number of tiles in y
     * @param imp The image stack to be processed
     * @return The new stack
     */
    static public ImagePlus tileXYZ(int nX, int nY, ImagePlus imp) {
        boolean isHyperStack = imp.isHyperStack();

        int nSlices, nFrames, nChannels;
        if (isHyperStack) {
            nSlices = imp.getNSlices();
            nFrames = imp.getNFrames();
            nChannels = imp.getNChannels();
        } else {
            nSlices = imp.getNSlices();
            nFrames = 1;
            nChannels = 1;
        }

        int dX = imp.getWidth();
        int dY = imp.getHeight();

        ImageStack resIms = new ImageStack(nX * dX, nY * dY);
        ImagePlus tmpImp;

        // TODO: fully implement zero image padding
        int totalSliceNumber = nSlices * nFrames * nChannels;
        int curSliceNumber;
        for (int iC = 0; iC < nChannels; iC++) {
            for (int iF = 0; iF < nFrames; iF++) {
                for (int iZ = 0; iZ < nSlices / (nX * nY); iZ++) {

                    curSliceNumber = iC * nFrames * nSlices + iF * nSlices + iZ;
                    IJ.showProgress(curSliceNumber,
                            totalSliceNumber);
                    IJ.showStatus("Tiling stack.");

                    tmpImp = new ImagePlus("tmp", new FloatProcessor(nX * dX, nY * dY));
                    for (int iY = 0; iY < nY; iY++) {
                        for (int iX = 0; iX < nX; iX++) {

                            imp.setPosition(iC + 1, nX * iY + iX + 1, iF + 1);
                            IJ.run(imp, "Copy", "");

                            tmpImp.setRoi(iX * dX, iY * dY, dX, dY);
                            IJ.run(tmpImp, "Paste", "");
                        }
                    }
                    resIms.addSlice(tmpImp.getProcessor());
                }
            }
        }

        IJ.showStatus("Done.");
        if (isHyperStack) {
            return HyperStackConverter.toHyperStack(new ImagePlus(imp.getTitle() + " - tiled", resIms),
                    nChannels, nSlices, nFrames, "xyztc", "grayscale");
        } else {
            return (new ImagePlus(imp.getTitle() + " - tiled", resIms));
        }
    }

    public static ImagePlus untile(int nX, int nY, int dX, int dY, ImagePlus imp) {
        ImageStack resIms = new ImageStack(dX, dY);
        for (int iY = 0; iY < nY; iY++) {
            for (int iX = 0; iX < nX; iX++) {
                imp.setRoi(iX * dX, iY * dY, dX, dY);
                resIms.addSlice(new Duplicator().run(imp).getProcessor());
            }
        }
        return (new ImagePlus(imp.getTitle() + "_stack", resIms));
    }

    /**
     *
     * @param imgs
     * @return
     */
    public static ImagePlus imagesToStack(ImagePlus[] imgs) {
        int width = imgs[0].getWidth();
        int height = imgs[0].getHeight();
        if (imgs.length > 1) {
            for (int i = 1; i < imgs.length; i++) {
                if ((imgs[i].getWidth() != width) | (imgs[i].getHeight() != height)) {
                    System.out.println("Images do not have the same width/height.");
                    return null;
                }
            }
        }
        ImageStack is = new ImageStack(width, height);
        for (ImagePlus img : imgs) {
            ImageStack curIs = img.getImageStack();
            for (int j = 0; j < curIs.getSize(); j++)
                is.addSlice(curIs.getProcessor(j + 1));
        }
        ImagePlus resImp = new ImagePlus(imgs[0].getTitle() + "ff.", is);
        return resImp;
    }

    public String[] stackToImageFiles(ImagePlus img, String path, String filePrefix){
        int nSlices, nFrames, nChannels;
        nSlices   = img.getNSlices();
        nFrames   = img.getNFrames();
        nChannels = img.getNChannels();
        
        ImagePlus           tmpImg;
        ImageProcessor      tmpIP;
        String              suffix;
        String[]            fileNames      = new String[nSlices*nFrames*nChannels];
        int                 currentImageNo = 0;
        
        for(int i = 0; i < nChannels; i++){
            for(int j = 0; j < nFrames; j++){
                for(int k = 0; k < nSlices; k++){
                    img.setPosition(i+1, k+1, j+1);
                    tmpIP  = img.getProcessor();
                    
                    suffix = String.format("_c%03d_f%03d_s%03d", (i+1), (j+1), (k+1));
                    tmpImg = new ImagePlus( img.getTitle() + suffix,
                                            tmpIP);
                    fileNames[currentImageNo] = filePrefix + suffix + ".tif";
                    SaveImage.saveTIFF(tmpImg, path + File.separator + fileNames[currentImageNo++]);
                }
            }
        }
        return(fileNames);
    }
    
    
    /**
     * Returns array with the mean pixel value for each slice
     *
     * @param imp
     * @return
     */
    public static double[][][] sliceWiseMean(ImagePlus imp) {
        double[][][] means = new double[imp.getNSlices()][imp.getNFrames()][imp.getNChannels()];

        ImageConverter ic = new ImageConverter(imp);
        ic.convertToGray32();

        for (int i = 0; i < imp.getNSlices(); i++) {
            for (int j = 0; j < imp.getNFrames(); j++) {
                for (int k = 0; k < imp.getNChannels(); k++) {
                    imp.setPosition(k + 1, i + 1, j + 1);
                    means[i][j][k] = imp.getProcessor().getStatistics().mean;
                }
            }
        }

        return means;
    }

    /**
     * Normalizes all images in a stack independently, i.e. the mean of each
     * image is 1.0. During the process the image type is changes to 32bit
     * float.
     *
     * @author greilich
     * @param imp An ImagePlus object with N slices
     * @return ImagePlus with same number of slices, normalized
     */
    public static ImagePlus sliceWiseNormalization(ImagePlus imp) {
        float curMean;
        float[] curPixels;
        ImageConverter ic = new ImageConverter(imp);
        ic.convertToGray32();
        ImageStack ims = imp.getStack();
        ImageStack resIms = new ImageStack(imp.getWidth(), imp.getHeight());
        for (int i = 0; i < imp.getNSlices(); i++) {
            curMean = (float) ims.getProcessor(i + 1).getStatistics().mean;
            curPixels = (float[]) ims.getProcessor(i + 1).getPixels();
            for (int j = 0; j < curPixels.length; j++)
                curPixels[j] /= curMean;
            
            FloatProcessor newFP = new FloatProcessor(imp.getWidth(), imp.getHeight(), curPixels);
            resIms.addSlice(newFP);
        }
        return new ImagePlus(imp.getTitle() + "SliceWiseNormalized", resIms);
    }

    /**
     * Normalizes all images in a stack independently, i.e. the mean of each
     * image is 1.0. During the process the image type is changes to 32bit
     * float.
     *
     * @param imp
     * @param refSliceIndex reference slice index, if zero (or out of bounds)
     * normalization will be to mean of stack
     * @return
     */
    public static ImagePlus stackWiseNormalization(ImagePlus imp, int refSliceIndex) {

        if (refSliceIndex < 0 || refSliceIndex > imp.getNSlices())
            refSliceIndex = 0;

        boolean silent = (IJ.getInstance() == null);

        Duplicator dp = new Duplicator();
        ImagePlus resImp = dp.run(imp);

        ImageConverter ic = new ImageConverter(resImp);
        ic.convertToGray32();

        int nChannels = resImp.getNChannels();
        int nFrames = resImp.getNFrames();
        int nSlices = resImp.getNSlices();
        double stackMean, sliceMean;
        for (int k = 0; k < nChannels; k++) {
            for (int j = 0; j < nFrames; j++) {
                if (!silent)
                    IJ.showStatus("Processing stack " + ((k + 1) * nChannels + j + 1) + " of " + nChannels * nFrames);

                if (refSliceIndex == 0) {
                    stackMean = 0.0;
                    for (int i = 0; i < nSlices; i++) {
                        resImp.setPosition(k + 1, i + 1, j + 1);
                        stackMean += resImp.getProcessor().getStatistics().mean;
                    }
                    stackMean /= nSlices;
                } else {
                    resImp.setPosition(k + 1, refSliceIndex, j + 1);
                    stackMean = resImp.getProcessor().getStatistics().mean;
                }

                for (int i = 0; i < nSlices; i++) {
                    resImp.setPosition(k + 1, i + 1, j + 1);
                    sliceMean = resImp.getProcessor().getStatistics().mean;
                    if (sliceMean != 0.0) {
                        resImp.getProcessor().multiply(stackMean / sliceMean);
                    }
                }
            }
        }

        resImp.setTitle(imp.getTitle() + " - stackwise normalized");
        return resImp;
    }
}
