package fntd.tracking;

import fntd.FNTDexception;
import fntd.access.FNTDprogress;
import ij.IJ;
import ij.ImagePlus;
import ij.process.ImageProcessor;
import java.util.Collections;
import java.util.Vector;
import java.util.ArrayList;

/**
 * For extrapolation of the tracks in the FNTD into the cell layer.
 * Finds the entry and exit voxel for each volume along a extrapolated track.
 * 
 * @author Jasper J.M. Kouwenberg
 */
public class CellHitStatistics
{
    
    /**
     * Array with nuclei hit statistics (nuclei ID : total passed track length, number of tracks)
     */
    public Vector<double[]>[] nuclei_info;
    
    private final int[][][]   pixels;
    
    private final int[]       cell_stack_dim;
    
    private final ImagePlus   cell_stack;
    
    private final double[][]  tracks;
    
    private final double[]    voxel_dim;
    
    private final double[]    FNTD_surface;
    
    private final double[][]  WER_to_LET;
    
    private final double[]    cell_surface;
    
    private final int[]       FNTD_stack_dim;
    
    private final int         no_of_tracks;
    
    private final double      diff_no_tracks;
    
    private final boolean     draw;

    /**
     * Constructs a Real Cell Calculation CellHitStatistics
     * @param _cell_stack ImagePlus cell image
     * @param _tracks Array with track coefficients (ID, x_slope, y_slope, x_offset, y_offset, residual_WER)
     * @param _voxel_dim Array with voxel dimensions (x,y,z)
     * @param _FNTD_stack_dim Array FNTD image stack dimensions (width, height, slices)
     * @param _FNTD_surface Array with plane coefficients for the FNTD surface reflection from within the FNTD (x_slope, y_slope, z_offset)
     * @param _cell_surface Array with plane coefficients for the cell surface reflection (x_slope, y_slope, z_offset)
     * @param _WER_to_LET Array WER (um) to LET (keV/um) translation data (WER, LET)
     * @param _no_of_tracks Number of tracks required for reference dose
     * @param _draw Draw the tracks through the cell stack
     *
     */
    public CellHitStatistics(ImagePlus _cell_stack, double[][] _tracks, double[] _voxel_dim, int[] _FNTD_stack_dim, double[] _FNTD_surface, double[] _cell_surface,
            double[][] _WER_to_LET, double _no_of_tracks, boolean _draw) {
        ImageProcessor ip;
        
        cell_stack      = _cell_stack;
        tracks          = _tracks;
        voxel_dim       = _voxel_dim;
        FNTD_surface    = _FNTD_surface;
        WER_to_LET      = _WER_to_LET;
        cell_surface    = _cell_surface;
        FNTD_stack_dim  = _FNTD_stack_dim;
        no_of_tracks    = (int) Math.floor(_no_of_tracks);
        diff_no_tracks  = _no_of_tracks - no_of_tracks;
        draw            = _draw;
        
        cell_stack_dim  = cell_stack.getDimensions(); //width, height, nChannels, nSlices, nFrames
        
        pixels          = new int[cell_stack_dim[3]][cell_stack_dim[0]][cell_stack_dim[1]];
        for (int i = 0; i < cell_stack_dim[3]; i++) {
            cell_stack.setPosition(1, i + 1, 1);
            ip = cell_stack.getProcessor().convertToShort(false);
            
            // Little slow, replace by getPixels
            for (int j = 0; j < cell_stack_dim[0]; j++)
                for (int k = 0; k < cell_stack_dim[1]; k++)
                    pixels[i][j][k] = ip.getPixel(j + 1, k + 1);
            
        }
    }
    
    /**
     * Constructs a Simulation CellHitStatistics
     *
     */
    public CellHitStatistics(double[][] _tracks, double[] _voxel_dim, int[] _FNTD_stack_dim, double[] _FNTD_surface, double[][] _WER_to_LET) {
        cell_stack      = null;
        tracks          = _tracks;
        voxel_dim       = _voxel_dim;
        FNTD_surface    = _FNTD_surface;
        WER_to_LET      = _WER_to_LET;
        cell_surface    = null;
        FNTD_stack_dim  = _FNTD_stack_dim;
        no_of_tracks    = 0;
        diff_no_tracks  = 0;
        draw            = false;
        cell_stack_dim  = null;
        pixels          = null;
    }

    /**
     * Constructs a Simulation CellHitStatistics
     *
     */
    public CellHitStatistics(double[][] _WER_to_LET) {
        cell_stack      = null;
        tracks          = null;
        voxel_dim       = null;
        FNTD_surface    = null;
        WER_to_LET      = _WER_to_LET;
        cell_surface    = null;
        FNTD_stack_dim  = null;
        no_of_tracks    = 0;
        diff_no_tracks  = 0;
        draw            = false;
        cell_stack_dim  = null;
        pixels          = null;
    }
    
    /**
     * Main nuclei hit statistics calculation function
     * @param CHS Class instance
     * @param dose_cycles Number of dose cycles
     * @param border_distance Distance to border in pixels
     * @param iterations number of iterations
     * @return double[][] (energy[], tracklength[], no. of tracks[])
     * @throws fntd.FNTDexception
     */
    public static double[][] RealCellHit(CellHitStatistics CHS, int dose_cycles, int border_distance, int iterations) throws FNTDexception {
        return CHS.RealCellHit(dose_cycles, border_distance, iterations);
    }
    
    /**
     * Main nuclei hit statistics calculation function
     * @param dose_cycles Number of dose cycles
     * @param border_distance Distance to border in pixels
     * @param iterations number of iterations
     * @throws fntd.FNTDexception
     */
    private double[][] RealCellHit(int dose_cycles, int border_distance, int iterations) throws FNTDexception {
        int x1, y1, z1, x2, y2, z2, pixelvalue, border_index;
        int[] indices, cell_offset_int;
        
        Vector<int[]> voxels, nucleus_voxels = new Vector<>(), cytoplasm_voxels = new Vector<>();
        
        ImagePlus cell_draw = new ImagePlus(); ImageProcessor dp;
        if (draw) cell_draw = (ImagePlus) cell_stack.duplicate();
        
        double zF, xF, yF, xC_slope, yC_slope, xE, yE, zE, cytoplasm_length, nucleus_length;
        double[] nucleus_energy = new double[iterations], nucleus_tracks = new double[iterations], nucleus_track_length = new double[iterations];
        
        ArrayList<Integer> track_rng = new ArrayList<Integer>(tracks.length);
        for(int i = 0; i < tracks.length; i++)
            track_rng.add(i);
        
        for (int it = 0; it < iterations; it++) {
            nucleus_energy[it]       = 0;
            nucleus_tracks[it]       = 0;
            nucleus_track_length[it] = 0;
            
            for (int dose = 0; dose < dose_cycles; dose++) {
                cell_offset_int = new int[]{
                    (int) border_distance + (int) Math.round(Math.random()*(FNTD_stack_dim[0] - 2*border_distance - cell_stack_dim[0])),
                    (int) border_distance + (int) Math.round(Math.random()*(FNTD_stack_dim[1] - 2*border_distance - cell_stack_dim[1]))};

                if (Math.random() < diff_no_tracks)
                    indices = new int[no_of_tracks + 1];
                else
                    indices = new int[no_of_tracks];

                Collections.shuffle(track_rng);
                for (int i = 0; i < indices.length; i++)
                    indices[i] = track_rng.get(i);

                for (int ind : indices) {
                    cytoplasm_voxels.clear();
                    nucleus_voxels.clear();
                    border_index = 0;

                    // Calculate track intercept with FNTD surface plan (x = xF, y = yF)
                    // See Notes.docx for derivation
                    zF = (FNTD_surface[1] * tracks[ind][2] + FNTD_surface[2] * tracks[ind][3] + FNTD_surface[0]) /
                            (1 - FNTD_surface[1] * tracks[ind][0] - FNTD_surface[2] * tracks[ind][1]);
                    xF = tracks[ind][0] * zF + tracks[ind][2];
                    yF = tracks[ind][1] * zF + tracks[ind][3];

                    // Find surface location in cell image
                    // The slopes are calculated in the FNTD where z corresponds to the deepness
                    // The trajectories travel therefore along the positive z into the cell
                    zF = (cell_surface[1] * xF + cell_surface[2] * yF + cell_surface[0]);

                    // Correct trajectory slope for surface tilts
                    xC_slope = tracks[ind][0] + cell_surface[1];
                    yC_slope = tracks[ind][1] + cell_surface[2];

                    // Find endpoint of track at the end of the cell stack
                    zE = (cell_stack_dim[3] - 1) * voxel_dim[2];
                    xE = xC_slope * (zE - zF) + xF;
                    yE = yC_slope * (zE - zF) + yF;

                    // Translate start and end location to pixels
                    x1 = (int) Math.floor(xF / voxel_dim[0]);
                    y1 = (int) Math.floor(yF / voxel_dim[1]);
                    z1 = (int) Math.floor(zF / voxel_dim[2]);
                    x2 = (int) Math.ceil(xE  / voxel_dim[0]);
                    y2 = (int) Math.ceil(yE  / voxel_dim[1]);
                    z2 = (int) (cell_stack_dim[3] - 1);

                    // Find Bresenham path between start and endpoint
                    voxels = this.Bresenham3D(x1, y1, z1, x2, y2, z2);

                    // Check if track hits the nucleus, if so, use for further calculation
                    for (int[] voxel : voxels) {
                        // If voxel falls outside of the cell image, skip
                        if (voxel[0] < cell_offset_int[0]     || voxel[0] >= cell_offset_int[0] + cell_stack_dim[0] ||
                                voxel[1] < cell_offset_int[1] || voxel[1] >= cell_offset_int[1] + cell_stack_dim[1])
                            continue;

                        // Retrieve color value of voxel
                        pixelvalue = pixels[voxel[2]][voxel[0] - cell_offset_int[0]][voxel[1] - cell_offset_int[1]];

                        if (pixelvalue == 182 || pixelvalue == 201) { // 0 = bg, 182 = cytoplasm, 201 = nucleus
                            cytoplasm_voxels.add(voxel);

                            if (pixelvalue == 201) {
                                nucleus_voxels.add(voxel);
                                if (nucleus_voxels.size() == 1)
                                    border_index = cytoplasm_voxels.size();
                            }
                            if (draw) {
                                cell_draw.setPosition(1, voxel[2] + 1, 1);
                                dp = cell_draw.getProcessor();
                                dp.setColor(pixelvalue == 182 ? 255 : 50);
                                dp.drawDot(voxel[0] + 1 - cell_offset_int[0], voxel[1] + 1 - cell_offset_int[1]);
                            }
                        }
                    }
                    if (nucleus_voxels.size() > 0) {
                        cytoplasm_length = nucleus_length = 0;
                        //Calculate tracklength through cytoplasm before nucleus is hit
                        //Note that the track is extrapolated from beneath the cell
                        if (border_index > 1)
                            cytoplasm_length  = this.Pathlength(cytoplasm_voxels.firstElement(), cytoplasm_voxels.elementAt(border_index - 2), voxel_dim, xC_slope, yC_slope, xF, yF, zF);

                        //Calculate tracklength through nucleus
                        if (nucleus_voxels.size() > 0)
                            nucleus_length    = this.Pathlength(nucleus_voxels.firstElement(), nucleus_voxels.lastElement(), voxel_dim, xC_slope, yC_slope, xF, yF, zF);

                        //Calculate nucleus statistics
                        nucleus_energy[it]       += this.LET_integration(tracks[ind][4] - cytoplasm_length, nucleus_length);
                        nucleus_track_length[it] += nucleus_length;
                        nucleus_tracks[it]       += 1;
                    }
                }
            }
            
            if (draw) fntd.access.DisplayImage.show(cell_draw);
        }
        
        double[][] result = {nucleus_energy, nucleus_track_length, nucleus_tracks};
        
        return result;
    }
    
    /**
     * Main simulation function
     * 
     * @param CHS Class instance
     * @return cytoplasm volume
     */
    public static double[][] SimulateCellHit(CellHitStatistics CHS, double Bounding_box, double Rnuc, double Rcyt_min, double Rcyt_max, int iterations) {
        return CHS.SimulateCellHit(Bounding_box, Rnuc, Rcyt_min, Rcyt_max, iterations);
    }
    
    /**
     * Main simulation function
     * 
     * @param cell_stack ImagePlus object of the cell layer
     * @return cytoplasm volume
     */
    private double[][] SimulateCellHit(double Bounding_box, double Rnuc, double Rcyt_min, double Rcyt_max, int iterations) {
        double[] r_to_nuc = new double[iterations]; 
        double[] r_in_nuc = new double[iterations]; 
        double[] energy   = new double[iterations]; 
        double[] o        = new double[3];
        double[] c        = new double[3];
        double[] l        = new double[3];
        double[] d        = new double[2];
        int      ind      = 0;
        double   norm, de;
        
        ArrayList<Integer> track_rng = new ArrayList<Integer>(tracks.length);
        for(int i = 0; i < tracks.length; i++)
            track_rng.add(i);
        
        int p = 0;
        
        while (p < iterations) {
            // Nucleus center
            c[0] = FNTD_stack_dim[0]*voxel_dim[0]/2.0 + (2.0*Math.random() - 1.0)*(Bounding_box - Rnuc);  //x
            c[1] = FNTD_stack_dim[1]*voxel_dim[1]/2.0 + (2.0*Math.random() - 1.0)*(Bounding_box - Rnuc);  //y
            c[2] = (Rcyt_min + Rnuc) + Math.random()*(Rcyt_max - Rcyt_min);                               //z
            
            // Collections.shuffle(track_rng);
            for (int i = 0; i < tracks.length; i++) {
                // ind = track_rng.get(i);
                ind = (int) Math.round(Math.random()*(tracks.length - 1));
                
                // Particle point of origin
                o[0] = tracks[ind][2] - c[0];
                o[1] = tracks[ind][3] - c[1];
                o[2] = 0              - c[2];
                
                if (Math.sqrt(o[0]*o[0] + o[1]*o[1]) > 5*Rnuc)
                    continue;
                
                // Particle direction
                l[0] = tracks[ind][0];
                l[1] = tracks[ind][1];
                l[2] = 1;
                
                norm = Math.sqrt(this.Dot_product(l, l));
                l[0] = l[0]/norm;
                l[1] = l[1]/norm;
                l[2] = l[2]/norm;
                
                de = 4.0*Math.pow(this.Dot_product(l, o), 2) - 4.0*(this.Dot_product(o, o) - Math.pow(Rnuc, 2));
                
                if (de <= 0)
                    continue;
                
                //Find intersects
                d[0] = -1.0*this.Dot_product(l, o) - Math.sqrt(de)/2.0;
                d[1] = -1.0*this.Dot_product(l, o) + Math.sqrt(de)/2.0;
                
                if (d[0] < 0 || d[1] < 0 || d[0] > tracks[ind][4])
                    continue;
                
                r_to_nuc[p] = d[0];
                if (d[1] > tracks[ind][4])
                    r_in_nuc[p] = tracks[ind][4] - d[0];
                else
                    r_in_nuc[p] = d[1] - d[0];

                energy[p] = this.LET_integration(tracks[ind][4] - d[0], r_in_nuc[p]);

                p += 1;

                if (IJ.getInstance() == null)
                    FNTDprogress.updateProgress((double) p / iterations);
                
                break;
            }
        }
        
        double[][] result = {r_to_nuc, r_in_nuc, energy};
        return result;
    }
    
    /**
     * Main simulation function
     * 
     * @param CHS Class instance
     * @return cytoplasm volume
     */
    public static double[][] SimulateSpheroid(CellHitStatistics CHS, double Bounding_box, int mode, double[] iso_range, double[] iso_prob, double Rnuc, double Reff, int iterations) {
        return CHS.SimulateSpheroid(Bounding_box, mode, iso_range, iso_prob, Rnuc, Reff, iterations);
    }
    
    /**
     * Main simulation function
     * 
     * @param cell_stack ImagePlus object of the cell layer
     * @return cytoplasm volume
     */
    private double[][] SimulateSpheroid(double Bounding_box, int mode, double[] iso_range, double[] iso_prob, double Rnuc, double Reff, int iterations) {
        double[] r_to_nuc = new double[iterations]; 
        double[] r_in_nuc = new double[iterations]; 
        double[] e_nuc    = new double[iterations]; 
        double[] e_cyt    = new double[iterations]; 
        double[] iso_ids  = new double[iterations]; 
        double[] px       = new double[iterations]; 
        double[] py       = new double[iterations]; 
        double[] pz       = new double[iterations]; 
        double[] no_cyt   = new double[iterations]; 
        double[] o        = new double[3];
        double[] d        = new double[2];
        double[] x        = new double[27];
        double[] y        = new double[27];
        double[] z        = new double[27];
        double[] l        = new double[3];
        double[] l_temp   = new double[2];
        int[]    ind      = new int[3];
        double de, iso;
        int iso_id;
        
        //MODE
        // 1: Random position in nucleus
        // 2: Random position in cytoplasm
        // 3: Random position on nucleus shell
        // 4: Random position on cytoplasm shell
        
        for (int i = 0; i < iterations; i++) {
            r_to_nuc[i] = 0;
            r_in_nuc[i] = 0;
            e_nuc[i]    = 0;
            e_cyt[i]    = 0;
            iso_ids[i]  = 0;
            no_cyt[i]   = 0;
        }
        
        int p = 0;
        while (p < iterations) {
            
            // Get random particle origin location outside of cell nuclei
            if (mode == 1 || mode == 2) {
                while (true) {
                    o[0] = (1.0 - 2.0*Math.random())*Bounding_box;
                    o[1] = (1.0 - 2.0*Math.random())*Bounding_box;
                    o[2] = (1.0 - 2.0*Math.random())*Bounding_box;

                    // Find closest cell nucleus center
                    for (int i = 0; i < 3; i++)
                        ind[i] = (int) Math.round(o[i] / (Math.sqrt(2) * Reff));

                    // Nuclei coordinates
                    int c = 0;
                    for (int i = 0; i < 27; i++) {
                        x[i] = 0;
                        y[i] = 0;
                        z[i] = 0;
                    }

                    // Get coordinates of 8 cell nuclei around the closest cell nucleus
                    for(int j = ind[0] - 1; j <= (ind[0] + 1); j++)
                        for(int k = ind[1] - 1; k <= (ind[1] + 1); k++)
                            for(int i = ind[2] - 1; i <= (ind[2] + 1); i++) {
                                if (Math.abs(i) % 2 == 0) { //even layer
                                    if ((Math.abs(j) % 2 == 0 && Math.abs(k) % 2 == 0) || (Math.abs(j) % 2 == 1 && Math.abs(k) % 2 == 1)) {
                                        x[c] = (double) j * Math.sqrt(2) * Reff;
                                        y[c] = (double) k * Math.sqrt(2) * Reff;
                                        z[c] = (double) i * Math.sqrt(2) * Reff;
                                    }
                                } else if (Math.abs(i) % 2 == 1) { //uneven layer
                                    if ((Math.abs(j) % 2 == 1 && Math.abs(k) % 2 == 0) || (Math.abs(j) % 2 == 0 && Math.abs(k) % 2 == 1)) {
                                        x[c] = (double) j * Math.sqrt(2) * Reff;
                                        y[c] = (double) k * Math.sqrt(2) * Reff;
                                        z[c] = (double) i * Math.sqrt(2) * Reff;
                                    }
                                }
                                c++;
                            }

                    // Check if the particle origin is inside one of the 9 closest cell nuclei
                    boolean fl_nuc = (mode == 2);
                    for (int i = 0; i < 27; i++) {
                        if (Math.sqrt(Math.pow(o[0] - x[i], 2) + Math.pow(o[1] - y[i], 2) + Math.pow(o[2] - z[i], 2)) <= Rnuc) {
                            fl_nuc = (mode == 1);
                            break;
                        }
                    }

                    if (fl_nuc)
                        break;
                }
            } else if (mode == 3 || mode == 4) {
                while (true) { // Select random nucleus
                    for (int i = 0; i < 3; i++)
                        ind[i] = (int) Math.round((1.0 - 2.0*Math.random())*Bounding_box / (Math.sqrt(2) * Reff));

                    if (Math.abs(ind[2]) % 2 == 0 && 
                            ((Math.abs(ind[0]) % 2 == 0 && Math.abs(ind[1]) % 2 == 0) || (Math.abs(ind[0]) % 2 == 1 && Math.abs(ind[1]) % 2 == 1))) //even layer
                        break;
                    else if (Math.abs(ind[2]) % 2 == 1 && 
                            ((Math.abs(ind[0]) % 2 == 1 && Math.abs(ind[1]) % 2 == 0) || (Math.abs(ind[0]) % 2 == 0 && Math.abs(ind[1]) % 2 == 1))) //uneven layer
                        break;
                }

                double radius;
                if (mode == 3)
                    radius = Rnuc;
                else
                    radius = Reff;

                // Random location on shell
                l_temp[0] = Math.random()*Math.PI;
                l_temp[1] = Math.random()*2*Math.PI;

                o[0]      = radius*Math.sin(l_temp[0])*Math.cos(l_temp[1]) + ind[0]*(Math.sqrt(2) * Reff);
                o[1]      = radius*Math.sin(l_temp[0])*Math.sin(l_temp[1]) + ind[1]*(Math.sqrt(2) * Reff);
                o[2]      = radius*Math.cos(l_temp[0]) + ind[2]*(Math.sqrt(2) * Reff);
            }

            // Grab a random direction vector
            l_temp[0] = Math.random()*Math.PI;
            l_temp[1] = Math.random()*2*Math.PI;
            l[0]      = Math.sin(l_temp[0])*Math.cos(l_temp[1]);
            l[1]      = Math.sin(l_temp[0])*Math.sin(l_temp[1]);
            l[2]      = Math.cos(l_temp[0]);
            
            // CYTOPLASM -----------------------------------------------------------------------------------
            // Check if the trajectory hits the center cytoplasm (https://en.wikipedia.org/wiki/Line%E2%80%93sphere_intersection)
            de = 4.0*Math.pow(this.Dot_product(l, o), 2) - 4.0*(this.Dot_product(o, o) - Math.pow(Reff, 2));
            if (de <= 0)
              continue;

            // Get nucleus-trajectory intersection points
            d[0] = -1.0*this.Dot_product(l, o) - Math.sqrt(de)/2.0;
            d[1] = -1.0*this.Dot_product(l, o) + Math.sqrt(de)/2.0;
            
            // Trajectory hits nucleus: select an isotope randomly
            iso = Math.random();
            iso_id = 0;
            for (int t = 0; t < iso_prob.length; t++)
                if (iso < iso_prob[t]) {
                    iso_id = t;
                    break;
                }

            // If the first intersection point is outside of the particle range, dismiss
            if (d[1] < 0 || d[0] > iso_range[iso_id])
                continue;
            
            // Calculate range in nucleus and energy deposition
            if (d[1] > iso_range[iso_id]) // Particle stopping inside cytoplasm
                e_cyt[p] += this.LET_integration(iso_range[iso_id] - d[0], iso_range[iso_id] - d[0]);
            else {
                if (Math.sqrt(Math.pow(o[0], 2) + Math.pow(o[1], 2) + Math.pow(o[2], 2)) <= Reff) // Particle originated inside cytoplasm
                    e_cyt[p] += this.LET_integration(iso_range[iso_id], d[1]);
                else // Regular passing through
                    e_cyt[p] += this.LET_integration(iso_range[iso_id] - d[0], d[1] - d[0]);
            }
            
            no_cyt[p]++;

            // NUCLEUS -------------------------------------------------------------------------------------
            // Check if the trajectory hits the center nucleus (https://en.wikipedia.org/wiki/Line%E2%80%93sphere_intersection)
            de = 4.0*Math.pow(this.Dot_product(l, o), 2) - 4.0*(this.Dot_product(o, o) - Math.pow(Rnuc, 2));
            if (de <= 0)
              continue;

            // Get nucleus-trajectory intersection points
            d[0] = -1.0*this.Dot_product(l, o) - Math.sqrt(de)/2.0;
            d[1] = -1.0*this.Dot_product(l, o) + Math.sqrt(de)/2.0;

            // If the first intersection points is outside of the particle range, dismiss
            if (d[1] < 0 || d[0] > iso_range[iso_id])
                continue;

            iso_ids[p]  = iso_id;
            r_to_nuc[p] = d[0];
            
            // Calculate range in nucleus and energy deposition
            if (d[1] > iso_range[iso_id])
                r_in_nuc[p] = iso_range[iso_id] - d[0];
            else {
                // Particle originated inside nucleus
                if (Math.sqrt(Math.pow(o[0], 2) + Math.pow(o[1], 2) + Math.pow(o[2], 2)) <= Rnuc)
                    r_in_nuc[p] = d[1];
                else
                    r_in_nuc[p] = d[1] - d[0];
            }
            
            e_nuc[p] = this.LET_integration(iso_range[iso_id] - d[0], r_in_nuc[p]);
            
            px[p] = o[0];
            py[p] = o[1];
            pz[p] = o[2];
            
            p += 1;

            if (IJ.getInstance() == null)
                FNTDprogress.updateProgress((double) p / iterations);
        }
        
        double[][] result = {iso_ids, r_to_nuc, r_in_nuc, e_nuc, e_cyt, px, py, pz, no_cyt};
        return result;
    }
    
    
    /**
     * Calculate pathlength using voxel wall entry and exit points
     * @param ivoxel_first First voxel
     * @param ivoxel_last Last voxel
     * @param x_slope x-z slope of track
     * @param y_slope y-z slope of track
     * @param x_intercept x intersect with cell surface
     * @param y_intercept y intersect with cell surface
     * @param z_intercept z intersect with cell surface
     */
    private double Pathlength(int[] ivoxel_first, int[] ivoxel_last, double[] voxel_dim, double x_slope, double y_slope, double x_intercept, double y_intercept, double z_intercept) {
        double o = 1;
        
        double[] voxel_first  = new double[ivoxel_first.length], voxel_last = new double[ivoxel_first.length];
        
        for (int i=0; i < ivoxel_first.length; i++)
            voxel_first[i] = ivoxel_first[i];
        
        for(int i=0; i < ivoxel_last.length; i++)
            voxel_last[i] = ivoxel_last[i];
            
        double z_start = voxel_first[2]*voxel_dim[2];      // z level of floor of first voxel 
        double z_end   = (voxel_last[2] + o)*voxel_dim[2]; // z level of ceilling of last voxel
        
        double x_start = (z_start - z_intercept)*x_slope + x_intercept;
        if (x_start < voxel_first[0]*voxel_dim[0])          x_start = voxel_first[0]*voxel_dim[0];
        if (x_start > (voxel_first[0] + o)*voxel_dim[0])    x_start = (voxel_first[0] + o)*voxel_dim[0];
        if (Math.abs(x_slope) > 0) z_start = (x_start - x_intercept)/x_slope + z_intercept;
        
        double y_start = (z_start - z_intercept)*y_slope + y_intercept;
        if (y_start < voxel_first[1]*voxel_dim[1])          y_start = voxel_first[1]*voxel_dim[1];
        if (y_start > (voxel_first[1] + o)*voxel_dim[1])    y_start = (voxel_first[1] + o)*voxel_dim[1];
        if (Math.abs(y_slope) > 0) z_start = (y_start - y_intercept)/y_slope + z_intercept;
        
        if (z_start < voxel_first[2]*voxel_dim[2] || z_start > (voxel_first[2]+o)*voxel_dim[2])
            z_start = voxel_first[2]*voxel_dim[2];
        
        // Calculate the trajectory length
        return Math.sqrt(Math.pow(z_end - z_start,2) * (1 + Math.pow(x_slope, 2) + Math.pow(y_slope, 2)));
    }
    
    /**
     * Calculates the deposited energy in keV using the energy table and the given ranges
     * 
     * @param residual_WER Residual range in water
     * @param nucleus_length Nucleus length
     * @return Energy
     */
    private double LET_integration(double residual_WER, double nucleus_length) {
        // LET table [Energy, LET, Range][row]
        double[] xvalues = {residual_WER, residual_WER - nucleus_length};
        if (xvalues[0] < 0) return 0;
        if (xvalues[1] < 0) xvalues[1] = 0;
        
        double[] yvalues = this.interpLinear(WER_to_LET[2], WER_to_LET[0], xvalues);
        
        return Math.abs(yvalues[0] - yvalues[1]);
    }
    
    /**
     * Calculates the volume of the nucleus
     * 
     * @param CHS Class instance
     * @return nucleus volume
     */
    public static double Nucleus_volume(CellHitStatistics CHS) {
        return CHS.Nucleus_volume();
    }
    
    /**
     * Calculates the volume of the nucleus
     * 
     * @param cell_stack ImagePlus object of the cell layer
     * @return nucleus volume
     */
    private double Nucleus_volume() {
        double pixelvalue, count = 0;
        ImageProcessor ip;
        
        for (int i = 1; i <= cell_stack_dim[3]; i++) {
            cell_stack.setPosition(1, i, 1);
            ip = cell_stack.getProcessor().convertToShort(false);
            for (int j = 1; j <= cell_stack_dim[0]; j++) {
                for (int k = 1; k <= cell_stack_dim[1]; k++) {
                    pixelvalue = ip.getPixel(j, k);
                    if (pixelvalue == 201)
                        count++;
                }
            }
                
        }
        
        return count*voxel_dim[0]*voxel_dim[1]*voxel_dim[2];
    }
    
    /**
     * Calculates the volume of the cytoplasm
     * 
     * @param CHS Class instance
     * @return cytoplasm volume
     */
    public static double Cytoplasm_volume(CellHitStatistics CHS) {
        return CHS.Cytoplasm_volume();
    }
    
    /**
     * Calculates the volume of the cytoplasm
     * 
     * @param cell_stack ImagePlus object of the cell layer
     * @return cytoplasm volume
     */
    private double Cytoplasm_volume() {
        double pixelvalue, count = 0;
        ImageProcessor ip;
        
        for (int i = 1; i <= cell_stack_dim[3]; i++) {
            cell_stack.setPosition(1, i, 1);
            ip = cell_stack.getProcessor().convertToShort(false);
            for (int j = 1; j <= cell_stack_dim[0]; j++) {
                for (int k = 1; k <= cell_stack_dim[1]; k++) {
                    pixelvalue = ip.getPixel(j, k);
                    if (pixelvalue == 182 || pixelvalue == 201)
                        count++;
                }
            }
                
        }
        
        return count*voxel_dim[0]*voxel_dim[1]*voxel_dim[2];
    }
    
    /**
     * Dot product
     * 
     * @param x Vector x
     * @param y Vector y
     * @return dot product
     */
    private double Dot_product(double[] x, double[] y) {
        double res = 0;
        for (int i = 0; i < x.length; i++)
            res += x[i]*y[i];
        
        return res;
    }
    
    /**
     * Bresenham function from https://gist.github.com/yamamushi/5823518
     * Original authorship accredited to Bob Pendelton
     * 
     * @param x1 Line start position x
     * @param y1 Line start position y
     * @param z1 Line start position z
     * @param x2 Line end position x
     * @param y2 Line end position y
     * @param z2 Line end position z
     */
    private Vector<int[]> Bresenham3D(int x1, int y1, int z1, int x2, int y2, int z2) {
        Vector<int[]> voxels = new Vector<>();
        
        int i, dx, dy, dz, l, m, n, x_inc, y_inc, z_inc, err_1, err_2, dx2, dy2, dz2;
        int[] pixel = new int[3];

        pixel[0] = x1;
        pixel[1] = y1;
        pixel[2] = z1;
        dx = x2 - x1;
        dy = y2 - y1;
        dz = z2 - z1;
        x_inc = (dx < 0) ? -1 : 1;
        l = Math.abs(dx);
        y_inc = (dy < 0) ? -1 : 1;
        m = Math.abs(dy);
        z_inc = (dz < 0) ? -1 : 1;
        n = Math.abs(dz);
        dx2 = l << 1;
        dy2 = m << 1;
        dz2 = n << 1;

        if ((l >= m) && (l >= n)) {
            err_1 = dy2 - l;
            err_2 = dz2 - l;
            for (i = 0; i < l; i++) {
                voxels.add(new int[]{pixel[0],pixel[1],pixel[2]});
                if (err_1 > 0) {
                    pixel[1] += y_inc;
                    err_1 -= dx2;
                }
                if (err_2 > 0) {
                    pixel[2] += z_inc;
                    err_2 -= dx2;
                }
                err_1 += dy2;
                err_2 += dz2;
                pixel[0] += x_inc;
            }
        } else if ((m >= l) && (m >= n)) {
            err_1 = dx2 - m;
            err_2 = dz2 - m;
            for (i = 0; i < m; i++) {
                voxels.add(new int[]{pixel[0],pixel[1],pixel[2]});
                if (err_1 > 0) {
                    pixel[0] += x_inc;
                    err_1 -= dy2;
                }
                if (err_2 > 0) {
                    pixel[2] += z_inc;
                    err_2 -= dy2;
                }
                err_1 += dx2;
                err_2 += dz2;
                pixel[1] += y_inc;
            }
        } else {
            err_1 = dy2 - n;
            err_2 = dx2 - n;
            for (i = 0; i < n; i++) {
                voxels.add(new int[]{pixel[0],pixel[1],pixel[2]});
                if (err_1 > 0) {
                    pixel[1] += y_inc;
                    err_1 -= dz2;
                }
                if (err_2 > 0) {
                    pixel[0] += x_inc;
                    err_2 -= dz2;
                }
                err_1 += dy2;
                err_2 += dx2;
                pixel[2] += z_inc;
            }
        }
        voxels.add(new int[]{pixel[0],pixel[1],pixel[2]});
        
        return voxels;
    }
    
    /**
    * Static methods for doing useful math
    *
    * @author  : brian
    * @version : 1.1
    *
    * The Monterey Bay Aquarium Research Institute (MBARI)provides this
    * documentation and code 'as is', with no warranty, express or
    * implied, of its quality or consistency. It is provided without support and
    * without obligation on the part of MBARI to assist in its use, correction,
    * modification, or enhancement. This information should not be published or
    * distributed to third parties without specific written permission from
    * MBARI.
    *
    * Copyright 2002 MBARI'
    * MBARI Proprietary Information. All rights reserved.
    *
    */
    
    private double[] interpLinear(double[] x, double[] y, double[] xi) throws IllegalArgumentException {
        if (x.length != y.length)
            throw new IllegalArgumentException("X and Y must be the same length");
        if (x.length == 1)
            throw new IllegalArgumentException("X must contain more than one value");
        
        double[] dx = new double[x.length - 1];
        double[] dy = new double[x.length - 1];
        double[] slope = new double[x.length - 1];
        double[] intercept = new double[x.length - 1];
        int loc;

        // Calculate the line equation (i.e. slope and intercept) between each point
        for (int i = 0; i < x.length - 1; i++) {
            dx[i] = x[i + 1] - x[i];
            
            if (dx[i] == 0)
                throw new IllegalArgumentException("X must be montotonic. A duplicate " + "x-value was found");
            if (dx[i] < 0)
                throw new IllegalArgumentException("X must be sorted");
            
            dy[i] = y[i + 1] - y[i];
            slope[i] = dy[i] / dx[i];
            intercept[i] = y[i] - x[i] * slope[i];
        }

        // Perform the interpolation here
        double[] yi = new double[xi.length];
        for (int i = 0; i < xi.length; i++) {
            if ((xi[i] > x[x.length - 1]) || (xi[i] < x[0]))
                yi[i] = Double.NaN;
            else {
                loc = java.util.Arrays.binarySearch(x, xi[i]);
                if (loc < -1) {
                    loc = -loc - 2;
                    yi[i] = slope[loc] * xi[i] + intercept[loc];
                }
                else
                    yi[i] = y[loc];
            }
        }

        return yi;
    }
    
}