package fntd.tracking;

import mosaic.core.detection.Particle;

/**
 * Extension of the MOSAIC Particle class. Contains extra variables that assist
 * with linking. Can be cast to MOSAIC Partice for limited backward compatibility
 * <p>
 * Redundant MOSAIC Particle variables:<br>
 *  - prev[], replaced by prev_slice and prev_index<br>
 *  - next[], replaced by next_slice and next_index<br>
 *  - special<br>
 * 
 * @author Jasper Kouwenberg
 */
public class FNTDParticle extends Particle {

    /**
     * Trajectory serial number this particle belongs to
     */
    public int trackNo = 0;
    
    /**
     * Contains the average velocity of this and all previous particles
     * in the trajectory.
     * Used to decrease prevent having to recalculate the average velocity
     * vector during linking of the particles.
     */
    public float lxh = 0, lyh = 0, lzh = 0;
    
    /**
     * The cost of the backward link.
     * Used to evaluate whether better links are available
     * when linkrange is larger then 1.
     */
    public float cost = Float.MAX_VALUE;
    
    /**
     * Indices of the channel, slice, frame and particle index of this particle
     * in the storage vector
     */
    public int channel = 0, slice = 0, frame = 0, index = 0;
    
    /**
     * Slice and particle index of the backward link
     */
    public int prev_slice = -1, prev_index = -1;
    
    /**
     * Slice and particle index of the forward link
     */
    public int next_slice = -1, next_index = -1;
    
    /**
     * Constructs a FNTDParticle from a MOSAIC Particle
     * 
     * @param p MOSAIC Particle object
     */
    public FNTDParticle(Particle p) {
        this(0, p.iX, p.iY, p.lx, p.ly, p.m0, p.m1, p.m2, p.m3, p.m4, p.nonParticleDiscriminationScore);
        this.setFrame(p.getFrame());
    }
    
    /**
     * Constructs a FNTDParticle from a MOSAIC Particle
     * 
     * @param p MOSAIC Particle object
     */
    public FNTDParticle(FNTDParticle p) {
        this(p.trackNo, p.iX, p.iY, p.lx, p.ly, p.m0, p.m1, p.m2, p.m3, p.m4, p.nonParticleDiscriminationScore);
        
        this.lxh        = p.lxh;
        this.lyh        = p.lyh;
        this.lzh        = p.lzh;
        this.cost       = p.cost;
        
        this.setFrame(p.getFrame());
        this.setIndex(p.channel, p.frame, p.slice, p.index);
        
        this.prev_index = p.prev_index;
        this.prev_slice = p.prev_slice;
        this.next_index = p.next_index;
        this.next_slice = p.next_slice;
    }
    
    /**
     * Constructs a FNTDParticle from given data
     * 
     * @param trackNo Track id
     * @param x x position
     * @param y y position
     * @param nonParticleDiscriminationScore Score
     * @param m0 Zero-order intensity momentum
     * @param m1 First-order intensity momentum
     * @param m2 Second-order intensity momentum
     * @param m3 Third-order intensity momentum
     * @param m4 Fourth-order intensity momentum
     * @param lx Velocity vector in x direction
     * @param ly Velocity vector in y direction
     */
    public FNTDParticle(int trackNo, double x, double y, double lx, double ly, double m0, double m1, double m2, double m3, double m4, double nonParticleDiscriminationScore) {
        this.trackNo = trackNo;
        this.iX  = (float) x;
        this.iY  = (float) y;
        this.lx = (float) lx;
        this.ly = (float) ly;
        this.m0 = (float) m0;
        this.m1 = (float) m1;
        this.m2 = (float) m2;
        this.m3 = (float) m3;
        this.m4 = (float) m4;
        this.nonParticleDiscriminationScore = (float) nonParticleDiscriminationScore;
    }
    
    /**
     * Sets the indices. Must be run after adding a particle
     * 
     * @param channel Channel number
     * @param frame Frame number
     * @param slice Slice number
     * @param index Particle index
     */
    public void setIndex(int channel, int frame, int slice, int index) {
        this.channel    = channel;
        this.frame      = frame;
        this.slice      = slice;
        this.index      = index;
    }
    
    /**
     * Unsets the linking parameters
     * 
     */
    public void toDefault() {
        this.trackNo = 0;
        this.cost = Float.MAX_VALUE;
        this.prev_index = -1;
        this.prev_slice = -1;
        this.next_index = -1;
        this.next_slice = -1;
        this.lxh = 0;
        this.lyh = 0;
        this.lzh = 0;
        this.lx = 0;
        this.ly = 0;
        this.lz = 0;
    }

    /**
     * Calculate the distance to particle p in x,y-frame
     *
     * @param p Target FNTDParticle
     * @return float
     */
    public float distance(FNTDParticle p) {
        return (float) Math.sqrt(this.distanceSq(p));
    }

    /**
     * Calculate the squared distance to particle p in x,y-frame
     *
     * @param p Target FNTDParticle
     * @return float
     */
    public float distanceSq(FNTDParticle p) {
        return (float) (Math.pow(p.iX - this.iX,2) + Math.pow(p.iY - this.iY,2));
    }

    /**
     * Get the magnitude of the 2D linking vector
     *
     * @return float
     */
    @Override
    public float linkModule() {
        return (float) Math.sqrt(this.linkModuleSq());
    }

    /**
     * Get the squared magnitude of the 2D linking vector
     *
     * @return float
     */
    public float linkModuleSq() {
        return lx * lx + ly * ly;
    }

    /**
     * Get the magnitude of the trajectory average linking factor till this particle
     *
     * @return float
     */
    public float linkModuleH() {
        return (float) Math.sqrt(this.linkModuleHSq());
    }

    /**
     * Get the square magnitude of the trajectory average linking factor till this particle
     *
     * @return float
     */
    public float linkModuleHSq() {
        return lxh * lxh + lyh * lyh;
    }
    
    public boolean match(Particle aParticle) {
        return iX == aParticle.iX && 
               iY == aParticle.iY &&
               iZ == aParticle.iZ &&
               m0 == aParticle.m0 &&
               m1 == aParticle.m1 &&
               m2 == aParticle.m2 &&
               m3 == aParticle.m3 &&
               m4 == aParticle.m4;
    }
}
