package fntd.tracking;

import fntd.access.FNTDprogress;
import fntd.tracking.FNTDtracks.Trajectory;
import ij.IJ;
import ij.ImagePlus;
import ij.process.ImageProcessor;
import java.util.Arrays;

import java.util.Vector;

/**
 * Class for particle linking and trajectory relinking
 * 
 * @author Jasper Kouwenberg
 */
public class FNTDParticleLinker {
    
    /**
     * Holds the current (maximum) linkrange. Is adjusted to stay within the
     * total number of slices.
     */
    int curr_linkrange;
    
    /**
     * Cost matrix. The first dimension represents the particles in the
     * current slice. The second dimension represents the cumulative particles
     * in the slices within the linkrange
     */
    float[][] C;
    
    /**
     * Association vector filled with the linking particles indices for PAn
     */
    int[] G;
            
    /**
     * Temporary value that holds the maximum allowed linking cost. Usually given
     * by ( 2 * max_displacement )^2
     */
    float max_cost;
    
    /**
     * Hold the current linking candidates
     */
    Vector<FNTDParticle> p1, p2;

    /**
     * Main function for linking particles in a stack of slices.
     * 
     * @param particles Array of FNTDParticle vectors
     * @param lo FNTDlinkerOptions
     * @return boolean
     */
    public boolean linkParticles(Vector<FNTDParticle>[] particles, FNTDlinkerOptions lo) {
        int n, i, j, k, count;
        float lx, ly, lx1, ly1, lx2, ly2, l1_m, l2_m;
        int dim_PAn, dim_PAn_k;
        FNTDParticle temp;
        
        this.curr_linkrange = lo.linkrange;
        this.max_cost       = (float) Math.pow(2*lo.displacement, 2);
        
        // Adjust linkrange to stay within bounds
        if (particles.length < (this.curr_linkrange + 1))
            this.curr_linkrange = particles.length - 1;

        for (n = 0; n < particles.length - this.curr_linkrange; n++) {
            IJ.showStatus("Linking slice " + (n + 1));
            dim_PAn = particles[n].size();

            // Calculate the particles that have been encountered at frame x.
            // Used to fill the second dimension of matrix g
            int[] acc_frame_part = new int[this.curr_linkrange + 1];
            acc_frame_part[0]    = 0;

            for (i = 1; i <= this.curr_linkrange; i++)
                acc_frame_part[i] = acc_frame_part[i - 1] + particles[n + i].size();

            dim_PAn_k = acc_frame_part[this.curr_linkrange];
            
            this.G = new int[dim_PAn];
            Arrays.fill(this.G, -1);

            this.C = new float[dim_PAn][dim_PAn_k];

            for (k = 0; k < this.curr_linkrange; k++) {
                this.p1 = particles[n];
                this.p2 = particles[n + (k + 1)];

                // Loop through the particles in slice n
                for (i = 0; i < dim_PAn; i++) {
                    // Loop through the particles in slice k
                    for (j = acc_frame_part[k]; j < acc_frame_part[k + 1]; j++) {
                        int p2_index = j - acc_frame_part[k];
                        float distance = this.p1.elementAt(i).distanceSq(p2.elementAt(p2_index));
                        
                        this.C[i][j] = (float) lo.l_d * (distance + k*k);

                        // Skip this link if the distance is too large
                        if (this.C[i][j] > this.max_cost)
                            continue;

                        // Calculate the cost based on the distance squared and the differences
                        // in intensity momenta
                        this.C[i][j] += lo.l_i / (lo.drift * lo.drift * Math.PI)
                                * (Math.pow(this.p1.elementAt(i).m0 - this.p2.elementAt(p2_index).m0, 2)
                                + Math.pow(this.p1.elementAt(i).m2 - this.p2.elementAt(p2_index).m2, 2));

                        // Skip this link if the cost is too large
                        if (this.C[i][j] > this.max_cost)
                            continue;

                        // If velocity feature is used and particle p1 has been linked before
                        if (lo.l_v > 0 && this.p1.elementAt(i).distance > 0) {
                            lx = (p2.elementAt(p2_index).iX - this.p1.elementAt(i).iX)/(k + 1);
                            ly = (p2.elementAt(p2_index).iY - this.p1.elementAt(i).iY)/(k + 1);

                            float f_magn_sq;
                            if (lo.angle_history > 1) 
                                f_magn_sq = (lx * lx + ly * ly) - this.p1.elementAt(i).linkModuleHSq();
                            else
                                f_magn_sq = (lx * lx + ly * ly) - this.p1.elementAt(i).linkModuleSq();
                            
                            
                            // Add the cost of the velocity differences squared
                            this.C[i][j] += lo.l_v * Math.abs(f_magn_sq);

                            // Skip this link if the cost is too large
                            if (this.C[i][j] > this.max_cost)
                                continue;
                        }
                        // If angle feature is used
                        if (lo.l_a > 0) {
                            if (lo.angle_history > 1) { // Use vector average velocity vectors
                                l1_m = this.p1.elementAt(i).linkModuleH();

                                lx1 = this.p1.elementAt(i).lxh/l1_m;
                                ly1 = this.p1.elementAt(i).lyh/l1_m;
                            } else { // Use velocity vectors of the last link
                                l1_m = this.p1.elementAt(i).linkModule();

                                lx1 = this.p1.elementAt(i).lx/l1_m;
                                ly1 = this.p1.elementAt(i).ly/l1_m;
                            }

                            if (l1_m > 0) { // If particle p1 has been linked before, ie has a velocity vector
                                // Set a standard cost if the distance is below particle radius / 2
                                // Allows for small variantes for particles with a small incidence angle
                                if (Math.sqrt(distance) <= lo.drift / 2 && l1_m <= lo.drift/2) {
                                    this.C[i][j] += lo.l_a * Math.pow((Math.cos(lo.max_angle * Math.PI / 180.0f) - 1) / 2 * lo.displacement, 2);
                                    continue;
                                }
                                
                                lx2 = this.p2.elementAt(p2_index).iX - this.p1.elementAt(i).iX;
                                ly2 = this.p2.elementAt(p2_index).iY - this.p1.elementAt(i).iY;

                                l2_m = (float) Math.sqrt(lx2 * lx2 + ly2 * ly2);
                                
                                // Normalize velocity vectors
                                lx2 /= l2_m;
                                ly2 /= l2_m;

                                // Calculate the cosine
                                float cos_phi = lx1 * lx2 + ly1 * ly2;

                                if (cos_phi >= Math.cos(lo.max_angle * Math.PI / 180.0f)) // If angle is smaller than maximum angle given by the user, add cost
                                    this.C[i][j] += lo.l_a * Math.pow((cos_phi - 1) / 2 * lo.displacement, 2);
                                else // Else, set cost in infinity
                                    this.C[i][j] = Float.MAX_VALUE;
                            } else
                                // A cost slightly larger than the cost for the maximum angle is added when this link would form a new trajectory
                                // as to favor extending existing trajectories compared to creating new ones
                                this.C[i][j] += lo.l_a * 1.2 * Math.pow((Math.cos(lo.max_angle * Math.PI / 180.0f) - 1) / 2 * lo.displacement, 2);
                        }
                        
                        // Check whether the particle p1 is attempting to link to has actually already been linked when due to linkrange is larger then 1.
                        // If so, set the cost for this link to infinity if the cost is higher than the existing link to the target particle
                        // as to prevent unneccessary overwriting of the link
                        if (this.p2.elementAt(p2_index).prev_index > -1 && this.C[i][j] > this.p2.elementAt(p2_index).cost)
                            this.C[i][j] = Float.MAX_VALUE;
                    }
                }
            }

            // Find the best link for all the particles within the linkrange (personal preference compared to
            // looking for the best link for each particle in slice m)
            for (j = 0; j < dim_PAn_k; j++) {
                IJ.showStatus("Linking Frame " + (n + 1));
                IJ.showProgress(j, dim_PAn_k);

                this.FindBestLink(j, dim_PAn, dim_PAn_k);
            }
            
            for (k = 0; k < this.curr_linkrange; k++) {
                this.p2 = particles[n + (k + 1)];

                for (i = 0; i < dim_PAn; i++) {
                    // Loop through the particles in slice n
                    for (j = acc_frame_part[k]; j < acc_frame_part[k + 1]; j++) {
                        int p2_index = j - acc_frame_part[k];
                        if (this.G[i] == j) { // These particles have been linked
                            if (this.p2.elementAt(p2_index).prev_slice >= 0) { // Particle p2 was already linked due to linkrange is larger then 1
                                particles[this.p2.elementAt(p2_index).prev_slice].elementAt(this.p2.elementAt(p2_index).prev_index).next_index = -1; // Unset the link information in the previous
                                particles[this.p2.elementAt(p2_index).prev_slice].elementAt(this.p2.elementAt(p2_index).prev_index).next_slice = -1; // linking partner of particle p2
                            }
                            
                            // Store the linking information
                            this.p1.elementAt(i).next_index         = p2_index;
                            this.p1.elementAt(i).next_slice         = n + (k + 1);
                            this.p2.elementAt(p2_index).prev_index  = i;
                            this.p2.elementAt(p2_index).prev_slice  = n;

                            this.p2.elementAt(p2_index).distance    = (float) Math.sqrt(this.p1.elementAt(i).distance(this.p2.elementAt(p2_index)));
                            this.p2.elementAt(p2_index).cost        = this.C[i][j];

                            if (lo.l_v > 0 || lo.l_a > 0) { // If velocity or angle features are used
                                this.p2.elementAt(p2_index).lx = (this.p2.elementAt(p2_index).iX - this.p1.elementAt(i).iX)/(k + 1);
                                this.p2.elementAt(p2_index).ly = (this.p2.elementAt(p2_index).iY - this.p1.elementAt(i).iY)/(k + 1);

                                if (lo.angle_history > 1) { // If vector averages are used, calculate the new averages
                                    count = 1;
                                    lx = 0;
                                    ly = 0;
                                    temp = this.p1.elementAt(i);
                                    
                                    while (count <= lo.angle_history) {
                                        if (temp.prev_slice < 0 || temp.prev_index < 0) // Continue while there are particles available in trajectory
                                            break;
                                        
                                        lx += temp.lx;
                                        ly += temp.ly;
                                        count++;

                                        temp = particles[temp.prev_slice].elementAt(temp.prev_index);
                                    }
                                    
                                    this.p2.elementAt(p2_index).lxh = (this.p2.elementAt(p2_index).lx + lx) / count;
                                    this.p2.elementAt(p2_index).lyh = (this.p2.elementAt(p2_index).ly + ly) / count;
                                }
                            }
                        }
                    }
                }
            }
            if (n == (particles.length - this.curr_linkrange - 1) && this.curr_linkrange > 1) // Adjust the linkrange to stay within bounds
                this.curr_linkrange--;
            
            if (IJ.getInstance() == null)
                FNTDprogress.updateProgress(((float) (n + 1)) / ((float) (particles.length - this.curr_linkrange)));
        }
        
        if (IJ.getInstance() == null)
            FNTDprogress.updateProgress(1);
        
        return true;
    }

    /**
     * Self-calling function that finds the optimal link for a particle. Can break
     * sub-optimal links and reassigning new links to the particles from the broken link.
     * 
     * @param part_index Index of the particle that is going to be linked
     * @param dim_PAn Number of particles in the current slice
     * @param dim_PAn_next Number of particles in the next slice
     */
    private void FindBestLink(int part_index, int dim_PAn, int dim_PAn_next)
    {
        double min = this.max_cost, old_cost = Float.MAX_VALUE;
        int best = -1, old_index = 0;

        // Loop through all the particles in frame m
        for (int i = 0; i < dim_PAn; i++) {
            // If the cost for forming this link is lower then the existing and the cost is lower
            // than the lowest cost found till this point
            if (this.C[i][part_index] < min && (this.G[i] < 0 || (this.G[i] > -1 && this.C[i][part_index] < this.C[i][this.G[i]]))) {
                best = i; // Set index of best link
                min = this.C[i][part_index]; // Set minimum cost to the cost of this link
            }
        }

        if (best == -1) // No link satisfying the above conditions has been found
            return;

        if (this.G[best] > -1) {
            old_cost  = this.C[best][this.G[best]];
            old_index = this.G[best];
        }

        this.G[best] = part_index;

        if (old_cost < Float.MAX_VALUE) // If this link overwrites an existing link
            this.FindBestLink(old_index, dim_PAn, dim_PAn_next); // Find next best link for unlinked particle
    }
    
    /**
     * Relinks tracks that have been interrupted using an approach similar to linkParticles.
     *
     * @param trajectories Vector of Trajectories
     * @param particles Vector of FNTDParticles
     * @param lo LinkerOptions
     */
    public static void reLinkTrajectories(Vector<Trajectory> trajectories, Vector<FNTDParticle>[] particles, FNTDlinkerOptions lo) {
        Vector<Trajectory> to_be_removed = new Vector<>();

        int min_particles = 3;

        int index, direction;
        double cos_phi, max_cost, cost;
        double l1_m, l2_m, lx1, ly1, lx2, ly2, l_x, l_y, l_tx, l_ty;
        double[] dist = new double[2];

        Vector<FNTDParticle> new_traj_particles;
        Trajectory new_trac, t1, t2;

        int total = trajectories.size();

        // Loop through the trajectories
        for (int l = 0; l < total; l++) {
            t1 = trajectories.elementAt(l);
            index = -1;
            max_cost = lo.relink_cost_factor * 4 * lo.relink_distance * lo.relink_distance; // Calculate the maximum allowed cost, based on the cost factor given by the user
            for (int m = l + 1; m < total; m++) { // Loop through the trajectories with a higher index than trajectory m, as to form a triangular matrix
                t2 = trajectories.elementAt(m);
                
                if (t1.existing_particles.size() < min_particles || t2.existing_particles.size() < min_particles) // Ignore trajectories that are too short
                    continue;

                dist[0] = Math.pow(t1.existing_particles.lastElement().iX - t2.existing_particles.firstElement().iX, 2) + Math.pow(t1.existing_particles.lastElement().iY - t2.existing_particles.firstElement().iY, 2) + Math.pow(t1.existing_particles.lastElement().iZ - t2.existing_particles.firstElement().iZ, 2);
                dist[1] = Math.pow(t1.existing_particles.firstElement().iX - t2.existing_particles.lastElement().iX, 2) + Math.pow(t1.existing_particles.firstElement().iY - t2.existing_particles.lastElement().iY, 2) + Math.pow(t1.existing_particles.firstElement().iZ - t2.existing_particles.lastElement().iZ, 2);

                if (Math.sqrt(dist[0]) > lo.relink_distance && Math.sqrt(dist[1]) > lo.relink_distance) // Check if either of the end-to-end distances are within the allowed displacement
                    continue;
                
                // Find the correct order of the trajectories based on the slice indices
                if (t2.start_slice - t1.stop_slice > 0 && t2.start_slice - t1.stop_slice <= 2 * lo.linkrange + 1)
                    direction = 0;
                else if (t1.start_slice - t2.stop_slice > 0 && t1.start_slice - t2.stop_slice < 2 * lo.linkrange + 1)
                    direction = 1;
                else
                    continue;

                l1_m = Math.sqrt(Math.pow(t1.lx, 2) + Math.pow(t1.ly, 2));
                l2_m = Math.sqrt(Math.pow(t2.lx, 2) + Math.pow(t2.ly, 2));

                lx1 = t1.lx / l1_m;
                ly1 = t1.ly / l1_m;
                lx2 = t2.lx / l2_m;
                ly2 = t2.ly / l2_m;

                cos_phi = lx1 * lx2 + ly1 * ly2;

                // Check whether the angle between the trajectory directions is within the user given limit
                if (Math.abs(Math.acos(cos_phi)) > lo.max_angle * Math.PI / 180)
                    continue;

                l_x = 0;
                l_y = 0;
                // Calculate the velocity vector between the connection points of the trajectories, based on the min_particles last particles in the first
                // trajectory and the min_particles first particles in the latter trajectory.
                for (int i = 0; i < min_particles; i++) {
                    if (direction == 0) {
                        l_tx = t2.existing_particles.elementAt(i).iX - t1.existing_particles.elementAt(t1.existing_particles.size() - (1 + i)).iX;
                        l_ty = t2.existing_particles.elementAt(i).iY - t1.existing_particles.elementAt(t1.existing_particles.size() - (1 + i)).iY;
                    } else {
                        l_tx = t1.existing_particles.elementAt(i).iX - t2.existing_particles.elementAt(t2.existing_particles.size() - (1 + i)).iX;
                        l_ty = t1.existing_particles.elementAt(i).iY - t2.existing_particles.elementAt(t2.existing_particles.size() - (1 + i)).iY;
                    }
                    l1_m = Math.sqrt(Math.pow(l_tx, 2) + Math.pow(l_ty, 2));

                    l_x += l_tx / (min_particles * l1_m);
                    l_y += l_ty / (min_particles * l1_m);
                }

                cos_phi = l_x * (lx1 + lx2) / 2 + l_y * (ly1 + ly2) / 2;

                // Check whether the connection vector between the trajectories aligns with the trajectory velocity vectors
                if (Math.abs(Math.acos(cos_phi)) > lo.max_angle * Math.PI / 180)
                    continue;

                cost =  lo.l_d * dist[direction];
                cost += lo.l_i / (lo.drift * lo.drift * Math.PI)
                        * (Math.pow(t1.m0 - t2.m0, 2)
                        + Math.pow(t1.m2 - t2.m2, 2));
                cost += lo.l_v * Math.abs(Math.pow(t1.lx, 2) + Math.pow(t1.ly, 2) - (Math.pow(t2.lx, 2) + Math.pow(t2.ly, 2)));
                cost += lo.l_a * Math.pow((cos_phi - 1) / 2 * lo.relink_distance, 2);
                
                // Check whether the above calculated cost is lower than the stored maximum cost
                if (cost < max_cost) {
                    index = m; // Store the index for trajectory
                    max_cost = cost; // Set the new maximum cost the cost of this relink
                }
            }

            if (index > -1) { // When a possible relink has been found
                t2 = trajectories.elementAt(index);
                
                new_traj_particles  = new Vector<>();
                new_trac            = new Trajectory(t2.serial_number);

                // Add the particles of the trajectories to the new trajectory in the correct order
                if (t2.start_slice - t1.stop_slice > 0 && t2.start_slice - t1.stop_slice <= 2 * lo.linkrange + 1) {
                    new_traj_particles.addAll(t1.existing_particles);
                    new_traj_particles.addAll(t2.existing_particles);
                } else {
                    new_traj_particles.addAll(t2.existing_particles);
                    new_traj_particles.addAll(t1.existing_particles);
                }
                
                for (FNTDParticle p : new_traj_particles)
                    new_trac.addParticle(p);

                // Replace the latter trajectory by the new trajectory. The latter trajectory will always appear later in the main for loop.
                // The new trajectory will therefore be checked for relinking later in the loop.
                trajectories.set(index, new_trac);

                // Add the first trajectory to the -to be removed- list
                to_be_removed.add(t1);
            }
        }

        // Remove the trajectories
        to_be_removed.stream().forEach((t) -> { 
            trajectories.remove(t);
        });
        
        FNTDParticle temp;
        for (Trajectory t : trajectories) {
            for (int i = 0; i < t.existing_particles.size(); i++) {
                // Rewrite the linking information in the particles based on the relinked trajectories
                temp = particles[t.existing_particles.elementAt(i).slice].elementAt(t.existing_particles.elementAt(i).index);
                
                temp.trackNo = t.serial_number;
                temp.prev_index = -1;
                temp.prev_slice = -1;
                temp.next_index = -1;
                temp.next_slice = -1;
                
                if (i > 0) {
                    temp.prev_index = t.existing_particles.elementAt(i-1).index;
                    temp.prev_slice = t.existing_particles.elementAt(i-1).slice;
                }
                if (i < (t.existing_particles.size() - 1)) {
                    temp.next_index = t.existing_particles.elementAt(i+1).index;
                    temp.next_slice = t.existing_particles.elementAt(i+1).slice;
                }
            }
        }
    }
    
    /**
     * Attempts to add particles to a track to estimate the end point more accurately
     *
     * @param imp ImagePlus
     * @param trajectories Vector of Trajectories
     * @param particles Vector array of FNTDParticles
     * @param radius Radius as used for particle selection
     */
    public static void expandTrackLength(ImagePlus imp, Vector<Trajectory> trajectories, Vector<FNTDParticle>[] particles, int radius) {
        FNTDParticle p_last, p_new;
        int px, py, pixel;
        double m0, m1, m2, m3, m4;
        ImageProcessor ipc;
        
        for (Trajectory t : trajectories) {
            if (t.existing_particles.lastElement().m0 > .25*t.m0 && t.existing_particles.lastElement().slice+2 < imp.getNSlices()) {
                p_last = t.existing_particles.lastElement();
                imp.setPosition(p_last.channel+1, p_last.frame+1, p_last.slice+2);
                ipc = imp.getProcessor();
                
                px = (int) Math.round(p_last.iX + t.lx);
                py = (int) Math.round(p_last.iY + t.ly);
                m0 = m1 = m2 = m3 = m4 = 0;
                
                for (int x = px - radius; x < py + radius; x++) {
                    if (x > 0 && x < imp.getWidth())
                        continue;
                    
                    for (int y = px - radius; y < py + radius; y++) {
                        if (y > 0 && y < imp.getHeight())
                            continue;
                    
                        pixel = ipc.getPixel(x, y);
                        
                        m0 += pixel;
                        m1 += ((x - px) + (y - py)) * pixel;
                        m2 += (Math.pow(x - px, 2) + Math.pow(y - py, 2)) * pixel;
                        m3 += (Math.pow(x - px, 3) + Math.pow(y - py, 3)) * pixel;
                        m4 += (Math.pow(x - px, 4) + Math.pow(y - py, 4)) * pixel;
                    }
                }
                
                p_new = new FNTDParticle(p_last.trackNo, px, py, p_last.lx, p_last.ly, m0, m1, m2, m3, m4, p_last.nonParticleDiscriminationScore);
                p_new.setIndex(p_last.channel, p_last.frame, p_last.slice+1, particles[p_last.slice+1].size());
                    
                p_new.prev_index = p_last.index;
                p_new.prev_slice = p_last.slice;
                p_last.next_index = p_new.index;
                p_last.next_slice = p_new.slice;
                
                particles[p_last.slice+1].addElement(p_new);
            }
        }
    }
    
}

