package fntd.tracking;

import fntd.FNTDexception;
import fntd.access.FNTDprogress;
import ij.IJ;
import ij.ImagePlus;
import ij.ImageStack;
import ij.gui.NewImage;
import ij.gui.OvalRoi;
import ij.gui.Roi;
import ij.plugin.Duplicator;
import ij.plugin.frame.RoiManager;
import ij.process.ImageProcessor;
import ij.process.ImageConverter;
import ij.process.ImageStatistics;
import ij.process.StackStatistics;
import trainableSegmentation.WekaSegmentation;
import ij.measure.ResultsTable;
import ij.plugin.filter.ParticleAnalyzer;
import java.awt.Rectangle;
import java.io.File;
import java.util.Arrays;
import java.util.Collections;
import java.util.Vector;
import mosaic.core.detection.FeaturePointDetector;
import mosaic.core.detection.MyFrame;
import mosaic.core.detection.Particle;
import org.apache.commons.lang3.ArrayUtils;

/**
 * General class for finding and linking of FNTDParticles
 *
 * @author Steffen Greilich
 */
public final class FNTDtracks {

    /**
     * Array of dimensions slices x frames x channels, where entry contains a
     * vector with the FNTDParticles in that slice.
     *
     * Order of dimensions: [channels][frames][slices]
     */
    public Vector<FNTDParticle> fntdParticles[][][];
    
    /**
     * Current highest track id
     */
    private int TrackNo = 0;
    
    /*************************************************************************
     * 
     * C O N S T R U C T O R S                                               *
     * 
     *************************************************************************/

    /**
     * Constructs empty FNTDtracks object
     */
    public FNTDtracks() {
        fntdParticles = null;
    }

    /**
     * Constructs an FNTDtracks object from the given object
     * 
     * @param ft FNTDObjects with particles
     */
    public FNTDtracks(FNTDtracks ft) {
        int nSlices   = ft.getNSlices();
        int nFrames   = ft.getNFrames();
        int nChannels = ft.getNChannels();
        
        this.fntdParticles = (Vector<FNTDParticle>[][][]) new Vector[nChannels][nFrames][nSlices];
        for (int i = 0; i < nChannels; i++)
            for (int j = 0; j < nFrames; j++)
                for (int k = 0; k < nSlices; k++) {
                    this.fntdParticles[i][j][k] = new Vector();
                    for (FNTDParticle p: ft.fntdParticles[i][j][k]) {
                        this.fntdParticles[i][j][k].add(new FNTDParticle(p));
                    }
                }
    }

    /**
     * Constructs an empty FNTDtracks object with same dimensions as the given image
     *
     * @param imp ImagePlus object (can be hypercube)
     */
    public FNTDtracks(ImagePlus imp) {
        fntdParticles = (Vector<FNTDParticle>[][][]) new Vector[imp.getNChannels()][imp.getNFrames()][imp.getNSlices()];
    }

    /**
     * Constructs FNTDtracks by calling the particle tracker
     * (FeaturePointExtraction) from the Mosaic tool suite
     *
     * @param imp ImagePlus object
     * @param rad Approx. radius of objects to be detected. This value is also used to calculate the intensity momenta.
     * @param cut Non-normalize cutoff value.
     * @param per Percentage of maximum intensity that is used for finding of particles.
     * @param threshold Minimum pixel value given as percentage of the maximum value in the stack
     * @throws FNTDexception General exception
     */
    public FNTDtracks(ImagePlus imp, int rad, double cut, double per, double threshold) throws FNTDexception {
        if (IJ.getInstance() == null)
            System.out.println("FINDING TRACKSPOTS");

        int nSlices = imp.getNSlices();
        int nFrames = imp.getNFrames();
        int nChannels = imp.getNChannels();

        StackStatistics stack_stats = new StackStatistics(imp);

        FeaturePointDetector fpd = new FeaturePointDetector((float) stack_stats.max, (float) stack_stats.min);

        if (threshold > 0)
            fpd.setDetectionParameters((float) cut, 0, rad, (float) (threshold/stack_stats.max), true);
        else
            fpd.setDetectionParameters((float) cut, (float) per / 100, rad, 0, false);

        int total = nSlices * nFrames * nChannels, particles = 0;
        double prog = 0;
        
        MyFrame[][][] myFrames = new MyFrame[nChannels][nFrames][nSlices];
        ImageStack ims;
        int nX = imp.getWidth(), nY = imp.getHeight();
        for (int i = 0; i < nChannels; i++) {
            for (int j = 0; j < nFrames; j++) {
                int stackIndex = 1;
                for (int k = 0; k < nSlices; k++) {
                    IJ.showStatus("Detecting slice " + (k + 1));
                    imp.setPosition(i + 1, k + 1, j + 1);
                    ims = new ImageStack(nX, nY);
                    ims.addSlice(imp.getProcessor());
                    ims.convertToFloat();
                    myFrames[i][j][k] = new MyFrame(ims, stackIndex++, 0);
                    myFrames[i][j][k].setParticles(fpd.featurePointDetection(ims));
                    
                    particles += myFrames[i][j][k].getParticles().size();

                    if (IJ.getInstance() == null)
                        FNTDprogress.updateProgress(prog / total);

                    prog++;
                }
            }
        }
        FNTDprogress.updateProgress(1);
        
        if (IJ.getInstance() == null)
            System.out.println("\n - Total spots found: " + particles + "\n");
        
        fromMyFrames(myFrames);
    }

    /**
     * Constructs FNTDtracks by calling the WEKA segmentation
     *
     * @param imp ImagePlus object (can be hypercube)
     * @param classifier Absolute path to the WEKA classifier
     * @param classificationImage Absolute path to the classification result image. 
     * The classification result is written to this path when either overwrite=true or when the image does not exist yet.
     * @param overwrite When true, forces classification of the input image, even when classificationImage points to an existing image.
     * @param radius The radius around the maximum pixel value that is used to calculate the intensity momenta.
     * @throws FNTDexception General exception
     */
    public FNTDtracks(ImagePlus imp, String classifier, String classificationImage, boolean overwrite, int radius) throws FNTDexception {
        ImagePlus imp_classified;

        // Check if there already exist a classification mask image and use it when overwrite = false
        if ((new File(classificationImage)).exists() == true && !overwrite)
            imp_classified = new ImagePlus(classificationImage);
        else
            imp_classified = this.WekaSegment(imp, classifier); // Call external function for image classification to ensure direct garbage collection

        int tot_part    = 0;
        int nX          = imp.getWidth();
        int nY          = imp.getHeight();
        int nSlices     = imp.getNSlices();
        int nFrames     = imp.getNFrames();
        int nChannels   = imp.getNChannels();
        int min_area    = (int) Math.floor(Math.PI / 4 * Math.pow(radius, 2));
        double pixel, diff_x, diff_y, m0, m1_x, m1_y;

        double ip_max = 0, ip_min = Double.MAX_VALUE;

        ImageStack          ims;
        Vector<Particle>    particles;
        Particle            temp_part;
        ImageProcessor      ip;
        Rectangle           rect;
        ParticleAnalyzer    analyzer;
        RoiManager          rois;
        
        // Create a new ParticleAnalyzer with a new RoiManager
        rois = new RoiManager();
        rois.setVisible(false);
        ParticleAnalyzer.setRoiManager(rois);
        analyzer
            = new ParticleAnalyzer(ParticleAnalyzer.ADD_TO_MANAGER | ParticleAnalyzer.INCLUDE_HOLES | ParticleAnalyzer.SHOW_NONE, 0, new ResultsTable(), min_area, 999);
        
        // Find the image maximum and minimum
        ip = imp.getProcessor();
        for (int i = 0; i < ip.getWidth(); i++) {
            for (int j = 0; j < ip.getHeight(); j++) {
                if (ip.getPixel(i, j) > ip_max)
                    ip_max = ip.getPixel(i, j);

                if (ip.getPixel(i, j) < ip_min)
                    ip_min = ip.getPixel(i, j);
            }
        }

        MyFrame[][][] myFrames = new MyFrame[nChannels][nFrames][nSlices];
        for (int i = 0; i < nChannels; i++) {
            for (int j = 0; j < nFrames; j++) {
                int stackIndex = 1;
                for (int k = 0; k < nSlices; k++) { // Loop through the slices
                    imp_classified.setPosition(i + 1, k + 1, j + 1);
                    
                    try { // Try to analyze the particles and store them in the RoiManager
                        analyzer.analyze(new ImagePlus("", imp_classified.getProcessor()));
                    } catch (NullPointerException e) {
                        throw new IllegalStateException("Rerunning this code gave a NullPointerException (Ya, screw JVM)", e);
                    }

                    // Store the current slice in the MyFrame
                    imp.setPosition(i + 1, k + 1, j + 1);
                    ip = imp.getProcessor();
                    ims = new ImageStack(nX, nY);
                    ims.addSlice(ip);
                    ims.convertToFloat();
                    myFrames[i][j][k] = new MyFrame(ims, stackIndex++, 0);

                    int n = rois.getCount();
                    particles = new Vector<>();
                    for (int l = 0; l < n; l++) { // Loop through the Roi's
                        rect = rois.getRoi(l).getBounds();
            
                        int max = 0, maxX = 0, maxY = 0;
                        // Find the location and value of the maximum value in the Roi
                        for (int x = (int) rect.getMinX(); x <= (int) rect.getMaxX(); x++) {
                            for (int y = (int) rect.getMinY(); y <= (int) rect.getMaxY(); y++) {
                                if (ip.getPixel(x, y) > max) {
                                    max = ip.getPixel(x, y);
                                    maxX = x;
                                    maxY = y;
                                }
                            }
                        }
                        
                        diff_x = diff_y = 1;
                        m0 = m1_x = m1_y = 0;
                        // Optimize particle center by intensity weighing, given by Crocker and Grier, 1996
                        while (diff_x >= 0.5 || diff_y >= 0.5) { // Loop while a significant change is present
                            for (int x = maxX - radius; x < maxX + radius; x++) {
                                if (x < 0 || x >= nX)
                                    continue;

                                for (int y = maxY - radius; y < maxY + radius; y++) {
                                    if (y < 0 || y >= nY)
                                        continue;

                                    // Calculate the intensity momenta of the found particle
                                    pixel = (ip.getPixel(x, y) - ip_min) / (ip_max - ip_min);
                                    m0   += pixel;
                                    m1_x += (x - maxX) * pixel;
                                    m1_y += (x - maxX) * pixel;
                                }
                            }
                            // Calculate intensity weighted difference from current center
                            diff_x = m1_x/m0;
                            diff_y = m1_y/m0;
                            
                            maxX   = (int) Math.round(maxX + diff_x);
                            maxY   = (int) Math.round(maxY + diff_y);
                        }

                        // Create a new MOSAIC particle at the location of the found maximum
                        temp_part = new Particle((float) maxX, (float) maxY, (float) k, j, 0);
                        temp_part.m0 = temp_part.m1 = temp_part.m2 = temp_part.m3 = temp_part.m4 = 0;

                        // Calculate intensity momenta
                        for (int x = maxX - radius; x < maxX + radius; x++) {
                            if (x < 0 || x >= nX)
                                continue;

                            for (int y = maxY - radius; y < maxY + radius; y++) {
                                if (y < 0 || y >= nY)
                                    continue;
                                
                                // Calculate the intensity momenta of the found particle
                                pixel = (ip.getPixel(x, y) - ip_min) / (ip_max - ip_min);
                                temp_part.m0 += pixel;
                                temp_part.m1 += ((x - maxX) + (y - maxY)) * pixel;
                                temp_part.m2 += (Math.pow(x - maxX, 2) + Math.pow(y - maxY, 2)) * pixel;
                                temp_part.m3 += (Math.pow(x - maxX, 3) + Math.pow(y - maxY, 3)) * pixel;
                                temp_part.m4 += (Math.pow(x - maxX, 4) + Math.pow(y - maxY, 4)) * pixel;
                            }
                        }

                        particles.add(temp_part);
                    }
                    
                    // Store the MOSAIC particles in the MyFrame
                    myFrames[i][j][k].setParticles(particles);

                    tot_part += particles.size();

                    rois.reset();
                }
            }
        }
        if (IJ.getInstance() == null)
            System.out.println("Total spots found: " + tot_part);

        fromMyFrames(myFrames); // Convert the MyFrames to FNTDParticles

        if ((new File(classificationImage)).exists() == false || overwrite)
            IJ.saveAsTiff(imp_classified, classificationImage);
    }

    /**
     * WEKA segmentation helper for the Construct using WEKA Segmentation. This routine is seperated and
     * put in a function to force garbage collection.
     *
     * @param imp ImagePlus object (can be hypercube)
     * @param classifier Absolute path to the WEKA classifier
     * 
     * @return ImagePlus Classified image mask
     */
    private ImagePlus WekaSegment(ImagePlus imp, String classifier) {
        // Classify image with given classifier
        WekaSegmentation segmentator = new WekaSegmentation(imp);
        segmentator.loadClassifier(classifier);
        segmentator.applyClassifier(1, false);
        ImagePlus imp_classified = segmentator.getClassifiedImage();

        // Convert classification mask to 8 bit for particle isolation
        ImageConverter conv = new ImageConverter(imp_classified);
        conv.convertToGray8();

        return imp_classified;
    }

    /**
     * Constructs an FNTDtracks object from an array of MyFrame objects as used
     * by the Mosaic tool suite
     *
     * @param myFrames MOSAIC MyFrames
     */
    public FNTDtracks(MyFrame myFrames[][][]) {
        fromMyFrames(myFrames);
    }

    /**
     * Extracts the FNTDParticles from the MOSAIC MyFrames
     * 
     * @param myFrames MOSAIC MyFrames
     */
    private void fromMyFrames(MyFrame[][][] myFrames) {
        int nChannels   = myFrames.length;
        int nFrames     = myFrames[0].length;
        int nSlices     = myFrames[0][0].length;
        
        fntdParticles = (Vector<FNTDParticle>[][][]) new Vector[nChannels][nFrames][nSlices];
        // Loop through stack in channels and frames
        for (int i = 0; i < nChannels; i++) {
            for (int j = 0; j < nFrames; j++) {
                for (int k = 0; k < nSlices; k++) {
                    fntdParticles[i][j][k] = new Vector<>();
                    
                    for (Particle p: myFrames[i][j][k].getParticles()) {
                        fntdParticles[i][j][k].add(new FNTDParticle(p));
                        fntdParticles[i][j][k].lastElement().setIndex(i, j, k, fntdParticles[i][j][k].size()-1);
                    }
                }
            }
        }
    }

    /**
     * Constructs FNTDtracks object from given track spots. Works for stacks yet
     * (no multi-frame, multi-channel images)
     *
     * @param slice Slice number
     * @param frame Frame number
     * @param channel Channel number
     * @param trackNo Track id
     * @param x x position
     * @param y y position
     * @param lx Velocity vector in x direction
     * @param ly Velocity vector in y direction
     * @throws FNTDexception General exception
     */
    public FNTDtracks(int[] slice, int[] frame, int[] channel, int[] trackNo, double[] x, double[] y, double[] lx, double[] ly) throws FNTDexception {
        double[] dummy = new double[trackNo.length];
        Arrays.fill(dummy, 0);
        fromData(slice, frame, channel, trackNo, x, y, lx, ly, dummy, dummy, dummy, dummy, dummy, dummy);
    }

    /**
     * Constructs FNTDtracks object from given track spots, including features
     * (m0-m4) and NPscor
     *
     * @param slice Slice number
     * @param frame Frame number
     * @param channel Channel number
     * @param trackNo Track id
     * @param x x position
     * @param y y position
     * @param nonParticleDiscriminationScore Score
     * @param m0 Zero-order intensity momentum
     * @param m1 First-order intensity momentum
     * @param m2 Second-order intensity momentum
     * @param m3 Third-order intensity momentum
     * @param m4 Fourth-order intensity momentum
     * @param lx Velocity vector in x direction
     * @param ly Velocity vector in y direction
     * @throws FNTDexception General exception
     */
    public FNTDtracks(int[] slice, int[] frame, int[] channel, int[] trackNo, double[] x, double[] y, double[] lx, double[] ly,
            double[] m0, double[] m1, double[] m2, double[] m3, double[] m4, double[] nonParticleDiscriminationScore) throws FNTDexception {
        fromData(slice, frame, channel, trackNo, x, y, lx, ly, m0, m1, m2, m3, m4, nonParticleDiscriminationScore);
    }

    /**
     * Constructs FNTDParticles using the given data
     * 
     * @param slice Slice number
     * @param frame Frame number
     * @param channel Channel number
     * @param trackNo Track id
     * @param x x position
     * @param y y position
     * @param nonParticleDiscriminationScore Score
     * @param m0 Zero-order intensity momentum
     * @param m1 First-order intensity momentum
     * @param m2 Second-order intensity momentum
     * @param m3 Third-order intensity momentum
     * @param m4 Fourth-order intensity momentum
     * @param lx Velocity vector in x direction
     * @param ly Velocity vector in y direction
     * @throws FNTDexception General exception 
     */
    private void fromData(int[] slice, int[] frame, int[] channel, int[] trackNo, double[] x, double[] y, double[] lx, double[] ly,
            double[] m0, double[] m1, double[] m2, double[] m3, double[] m4, double[] nonParticleDiscriminationScore) throws FNTDexception {

        int nChannels = 0;
        int nFrames = 0;
        int nSlices = 0;

        int n = slice.length;

        if (n == 0)
            throw new FNTDexception("No input data for FNTDtracks constructor routine...");

        if ((n != trackNo.length)
                || (n != x.length)
                || (n != y.length)
                || (n != lx.length)
                || (n != ly.length)
                || (n != channel.length)
                || (n != frame.length)) {
            throw new FNTDexception("Input array length mismatch in FNTDtracks constructor routine...");
        } else {
             // get number of channels, frames, slices
            for (int i = 0; i < n; i++) {
                nSlices = Math.max(slice[i], nSlices);
                nFrames = Math.max(frame[i], nFrames);
                nChannels = Math.max(channel[i], nChannels);
            }

            fntdParticles = (Vector<FNTDParticle>[][][]) new Vector[nChannels][nFrames][nSlices];

            // loop through FNTDtracks array and initialize
            for (int i = 0; i < nChannels; i++) {
                for (int j = 0; j < nFrames; j++) {
                    for (int k = 0; k < nSlices; k++) {
                        fntdParticles[i][j][k] = new Vector<>();
                        for (int l = 0; l < n; l++) {
                            if ((slice[l] == k + 1) && (frame[l] == j + 1) && (channel[l] == i + 1))
                                fntdParticles[i][j][k].add(new FNTDParticle(trackNo[l], x[l], y[l], lx[l], ly[l], m0[l], m1[l], m2[l], m3[l], m4[l], nonParticleDiscriminationScore[l]));
                        }
                    }
                }
            }
        }
    }

    /**
     * Duplicate the given object
     * 
     * @param ft FNTDtracks object
     * @return FNTDtracks
     * @throws FNTDexception General exception
     */
    public static FNTDtracks duplicate(FNTDtracks ft) throws FNTDexception {
        return (new FNTDtracks(ft));
    }

    
    /*************************************************************************
     * 
     * U T I L I T I E S                                                     *
     * 
     *************************************************************************/
    
    /**
     * Returns detailed information about the tracks
     * 
     * @param imp ImagePlus object
     * @param ft FNTDtracks object
     * @param radius Radius of the ROI around the maximum pixel value
     * @param criticalDistance Minimum distance between two spots
     * @param noBackgroundGauss if true, B will be set to zero for Gauss fit
     * @return Array with results, CAREFUL: Indices for frames, slices, channels
     * change from zero-based (Java style) to one-based (R or similar style)
     */
    public static double[][] getProperties(ImagePlus imp,
            FNTDtracks ft,
            double radius,
            double criticalDistance,
            boolean noBackgroundGauss) {
        double[][] result = ft.getProperties(imp, radius, criticalDistance, noBackgroundGauss);
        return (result);
    }

    /**
     * Returns detailed information about the trackspots
     * 
     * @param orgImp ImagePlus object
     * @param radius Radius of the ROI around the maximum pixel value
     * @param criticalDistance Minimum distance between two spots
     * @param noBackgroundGauss if true, B will be set to zero for Gauss fit
     * @return Array with results, CAREFUL: Indices for frames, slices, channels
     * change from zero-based (Java style) to one-based (R or similar style)
     */
    // TODO: LET THIS ROUTINE USE 'to.array()' TO MAKE SURE THAT TRACKSPOT
    // POSITION ETC. ARE DEFINITELY THE SAME.
    public double[][] getProperties(ImagePlus orgImp,
            double radius,
            double criticalDistance,
            boolean noBackgroundGauss) {
        
        System.out.println("CALCULATING TRACKSPOT PROPERTIES");

        int nSlices = orgImp.getNSlices();
        int nFrames = orgImp.getNFrames();
        int nChannels = orgImp.getNChannels();

        // Get a stack as copy of the original image
        // TODO: How to check if the order is actually ZTC?
        ImageStack is = orgImp.getImageStack();

        // Get ROIs from tracks and
        Vector[][][]     rois = toROIs(radius);
        boolean[][][][]  belowCriticalDistance = belowCriticalDistance(criticalDistance);

        final int nData = 9;
        double[][] allResults = new double[nData][];
        double[][] tmpResults = new double[nData][];
        double[][] newResults;
        ImageStatistics imageStatistics;
        ImageProcessor ip;

        Rectangle r;
        double[] x, y;
        int[] trackNo;
        double prog = 0, total = nChannels*nFrames*nSlices;

        for (int i = 0; i < nChannels; i++) {
            for (int j = 0; j < nFrames; j++) {
                for (int k = 0; k < nSlices; k++) {
                    if ((i >= rois.length) || (j >= rois[0].length) || (k >= rois[0][0].length))
                        continue;
                    
                    if (!rois[i][j][k].isEmpty()) {
                        ip = is.getProcessor((k + 1) + j * nSlices + i * (nSlices * nFrames));
                        x = getX(k, j, i);
                        y = getY(k, j, i);
                        trackNo = getTrackNo(k, j, i);
                        newResults = new double[nData][rois[i][j][k].size()];
                        
                        for (int l = 0; l < rois[i][j][k].size(); l++) {
                            newResults[0][l] = (double) i + 1;
                            newResults[1][l] = (double) j + 1;
                            newResults[2][l] = (double) k + 1;
                            newResults[3][l] = (double) trackNo[l];
                            newResults[4][l] = (double) x[l];
                            newResults[5][l] = (double) y[l];
                            
                            // If trackspot is not below critical distance to another, evaluate
                            if (belowCriticalDistance[i][j][k][l] == false) {
                                ip.setRoi((Roi)rois[i][j][k].get(l));
                                
                                // Get hottest pixel value (max) and mean over radius around center
                                imageStatistics = ip.getStatistics();
                                newResults[6][l] = imageStatistics.mean;
                                newResults[7][l] = imageStatistics.max;

                                // Get average of hottest pixel and its four adjacents
                                r = ((Roi)rois[i][j][k].get(l)).getBounds();
                                int[] center = {(int) Math.round(r.getCenterX()), (int) Math.round(r.getCenterY())};
                                newResults[8][l] = (ip.getPixelValue(center[0], center[1]) + ip.getPixelValue(center[0] - 1, center[1])
                                                    + ip.getPixelValue(center[0] + 1, center[1]) + ip.getPixelValue(center[0], center[1] - 1)
                                                    + ip.getPixelValue(center[0], center[1] + 1)) / 5;
                                
                            } else {
                                newResults[6][l] = 0.0;
                                newResults[7][l] = 0.0;
                                newResults[8][l] = 0.0;
                            }
                        }
                        for (int l = 0; l < nData; l++) {
                            if (allResults[l] != null) {
                                tmpResults[l] = new double[allResults[l].length + newResults[l].length];
                                System.arraycopy(allResults[l], 0, tmpResults[l], 0, allResults[l].length);
                                System.arraycopy(newResults[l], 0, tmpResults[l], allResults[l].length, newResults[l].length);
                                allResults[l] = tmpResults[l];
                            } else
                                allResults[l] = newResults[l];
                        }
                    }
                    FNTDprogress.updateProgress(prog / total);
                    prog++; 
                }
            }
        }
        FNTDprogress.updateProgress(1);
        System.out.println("");
        return allResults;
    }

    /**
     * Attempts to link the FNTDParticles and construct optimal trajectories
     * 
     * @param ft FNTDtracks object
     * @param linkrange Number of slices that are considered for linking to a particle in the current slice.
     * @param displacement Maximum displacement between FNTDParticles in the x-y direction, given in pixels.
     * @param drift Radius of the ROI around the maximum pixel value
     * @param WeightDistance Weight factor for the displacement in the cost function
     * @param WeightIntensity Weight factor for the intensity momenta difference in the cost function
     * @param WeightVelocity Weight factor for the velocity difference in the cost function
     * @param WeightAngle Weight factor for the angle difference in the cost function
     * @param max_angle Maximum angle allowed between the velocity vectors
     * @param angle_history Linking uses the average velocity vector over the whole track instead of 
     * only from the last link when set to true.
     * @param relink Attempts to relink interrupted trajectories when set to true.
     * @param relink_cost Maximum allowed trajectory relinking cost modifier. Usually between 0 and 1.
     * @param relink_distance Maximum allowed distance between the two closests endpoints of the trajectories.
     * @return FNTDtracks object
     * @throws FNTDexception General exception 
     */
    public static FNTDtracks linkParticles(FNTDtracks ft,
            int linkrange,
            int displacement,
            double drift,
            double WeightDistance,
            double WeightIntensity,
            double WeightVelocity,
            double WeightAngle,
            double max_angle,
            int angle_history,
            boolean relink,
            double relink_cost,
            double relink_distance) throws FNTDexception {

        ft.linkParticles(linkrange, displacement, drift, WeightDistance, WeightIntensity, WeightVelocity, WeightAngle,
                         max_angle, angle_history, relink, relink_cost, relink_distance);

        return (ft);
    }

    /**
     * Attempts to link the FNTDParticles and construct optimal trajectories
     * 
     * @param linkrange Number of slices that are considered for linking to a particle in the current slice.
     * @param displacement Maximum displacement between FNTDParticles in the x-y direction, given in pixels.
     * @param drift Radius of the ROI around the maximum pixel value
     * @param WeightDistance Weight factor for the displacement in the cost function
     * @param WeightIntensity Weight factor for the intensity momenta difference in the cost function
     * @param WeightVelocity Weight factor for the velocity difference in the cost function
     * @param WeightAngle Weight factor for the angle difference in the cost function
     * @param max_angle Maximum angle allowed between the velocity vectors
     * @param angle_history Linking uses the average velocity vector over the whole track instead of 
     * only from the last link when set to true.
     * @param relink Attempts to relink interrupted trajectories when set to true.
     * @param relink_cost Maximum allowed trajectory relinking cost modifier. Usually between 0 and 1.
     * @param relink_distance Maximum allowed distance between the two closests endpoints of the trajectories.
     * @return boolean
     */
    public boolean linkParticles(int linkrange,
            int displacement,
            double drift,
            double WeightDistance,
            double WeightIntensity,
            double WeightVelocity,
            double WeightAngle,
            double max_angle,
            int angle_history,
            boolean relink,
            double relink_cost,
            double relink_distance) {
        
        if (IJ.getInstance() == null)
            System.out.println("LINKING PARTICLES");

        FNTDlinkerOptions lo    = new FNTDlinkerOptions();
        lo.linkrange            = linkrange;
        lo.drift                = (float) drift;
        lo.displacement         = (float) displacement;
        lo.relink_cost_factor   = (float) relink_cost;
        lo.relink_distance      = (float) relink_distance;

        lo.l_d = (float) WeightDistance;
        lo.l_i = (float) WeightIntensity;
        lo.l_v = (float) WeightVelocity;
        lo.l_a = (float) WeightAngle;

        lo.max_angle     = (float) max_angle;
        lo.angle_history = angle_history;

        for (Vector<FNTDParticle>[][] f : fntdParticles)
            for (Vector<FNTDParticle>[] n : f)
                (new FNTDParticleLinker()).linkParticles(n, lo);
        
        this.updateTracks();
        
        if (IJ.getInstance() == null)
            System.out.println("\n - Total tracks found: " + this.getNTracks() + "\n");
        
        if (relink) {
            System.out.println("RELINKING TRACKS");
            IJ.showStatus("Relinking Tracks");
            
            int nChannels   = getNChannels();
            int nFrames     = getNFrames();
            int old_length  = this.getNTracks();
            double prog = 0;
            
            for (int i = 0; i < nChannels; i++) {
                for (int j = 0; j < nFrames; j++) {
                    FNTDprogress.updateProgress(prog / fntdParticles[0].length);
                    prog++;
                    
                    FNTDParticleLinker.reLinkTrajectories(this.toTrajectories(i, j), this.fntdParticles[i][j], lo);
                }
            }
            FNTDprogress.updateProgress(1);
            
            this.updateTracks();
            
            System.out.println("\n - Tracks combined: " + (old_length - this.getNTracks()) + "\n");
        }
        
        return true;
    }
    
    /**
     * Adds paricle when the last particle m0 is higher than 1/2 of the track average
     * 
     * @param imp ImagePlus object
     * @param ft FNTDtracks object
     * @param radius Radius of the ROI around the maximum pixel value
     * @return FNTDtracks
     * @throws fntd.FNTDexception
     */
    public static FNTDtracks expandTrackLength(ImagePlus imp, FNTDtracks ft, double radius) throws FNTDexception {
        ft.expandTrackLength(imp, radius);

        return (ft);
    }
    
    /**
     * Adds particle when the last particle m0 is higher than 1/2 of the track average
     * 
     * @param imp ImagePlus object
     * @param radius Radius of the ROI around the maximum pixel value
     */
    private void expandTrackLength(ImagePlus imp, double radius) {
        int nChannels = getNChannels();
        int nFrames = getNFrames();
        int length0, length1;

        for (int i = 0; i < nChannels; i++) {
            for (int j = 0; j < nFrames; j++) {
                length0 = 0; length1 = 0;
                
                for (Vector<FNTDParticle> fntdParticle : this.fntdParticles[i][j])
                    length0 += fntdParticle.size();
                
                FNTDParticleLinker.expandTrackLength(imp, this.toTrajectories(i, j), this.fntdParticles[i][j], (int) radius);
                
                for (Vector<FNTDParticle> fntdParticle : this.fntdParticles[i][j])
                    length1 += fntdParticle.size();
                
                System.out.println("Particles added: " + (length1 - length0));
            }
        }
    }
    
    /**
     * Updates track indices of particles within tracks. Used to recalculate the total number
     * of tracks and to make sure all particles are connected correctly.
     * 
     */
    public void updateTracks() {
        int nChannels = getNChannels();
        int nFrames = getNFrames();
        int nSlices = getNSlices();
        
        for (int i = 0; i < nChannels; i++)
            for (int j = 0; j < nFrames; j++)
                for (int k = 0; k < nSlices; k++)
                    for (FNTDParticle p : this.fntdParticles[i][j][k])
                        p.trackNo = 0;
        
        this.TrackNo = 1;
        for (int i = 0; i < nChannels; i++) {
            for (int j = 0; j < nFrames; j++) {
                for (int k = 0; k < nSlices; k++) {
                    for (FNTDParticle p : this.fntdParticles[i][j][k]) {
                        
                        if (p.prev_index > -1 && p.prev_slice > -1) {
                            p.trackNo = this.fntdParticles[i][j][p.prev_slice].elementAt(p.prev_index).trackNo;
                        } else if (p.next_index > -1 && p.next_slice > -1) {
                            p.trackNo = this.TrackNo;
                            this.TrackNo++;
                        } else {
                            p.trackNo = 0;
                        }
                    }
                }
            }
        }
    }
    
    /**
     * Resets all tracks but leaves the particles intact. Used in the Fiji plugin for
     * fast iterations of the linking parameters
     * 
     */
    public void resetTracks() {
        int nChannels = getNChannels();
        int nFrames = getNFrames();
        int nSlices = getNSlices();
        
        for (int i = 0; i < nChannels; i++)
            for (int j = 0; j < nFrames; j++)
                for (int k = 0; k < nSlices; k++)
                    for (FNTDParticle p : this.fntdParticles[i][j][k])
                        p.toDefault();
    }

    /**
     * Removes tracks that contain less than the given amount of FNTDParticles
     * 
     * @param ft FNTDtracks object
     * @param minLength Minimum number of particles in a track
     * @return FNTDtracks object
     * @throws FNTDexception General exception
     */
    public static FNTDtracks filterTrackLength(FNTDtracks ft, int minLength) throws FNTDexception {
        ft.filterTrackLength(minLength);

        return ft;
    }

    /**
     * Removes tracks that contain less than the given amount of FNTDParticles
     * 
     * @param minLength Minimum number of particles in a track
     * @throws FNTDexception General exception 
     */
    public void filterTrackLength(int minLength) throws FNTDexception {
        if (IJ.getInstance() == null)
            System.out.println("FILTERING TRACKS");
        
        int nChannels = getNChannels();
        int nFrames   = getNFrames();
        int nSlices   = getNSlices();
        
        int nTracks         = this.getNTracks();
        int[] track_lengths = new int[nTracks];
        double prog = 0, total = nChannels*nFrames*nSlices*2;

        // First count the track lenghts and the total number of particles inside tracks 
        // that are longer than the minimum length
        for (int i = 0; i < nChannels; i++) {
            for (int j = 0; j < nFrames; j++) {
                for (int k = 0; k < nSlices; k++) {
                    for (FNTDParticle p : this.fntdParticles[i][j][k]) {
                        if (p.trackNo < 1 || p.trackNo > nTracks)
                            continue;

                        track_lengths[p.trackNo - 1]++;
                    }
                    
                    FNTDprogress.updateProgress(prog/total);
                    prog++;
                }
            }
        }

        for (int i = 0; i < nChannels; i++) {
            for (int j = 0; j < nFrames; j++) {
                for (int k = 0; k < nSlices; k++) {
                    for (FNTDParticle p : this.fntdParticles[i][j][k]) {
                        if (p.trackNo == 0 || track_lengths[p.trackNo - 1] < minLength || p.trackNo > nTracks)
                        {
                            p.trackNo = 0;
                            p.prev_index = -1;
                            p.prev_slice = -1;
                            p.next_index = -1;
                            p.next_slice = -1;
                        }
                    }
                    
                    FNTDprogress.updateProgress(prog/total);
                    prog++;
                }
            }
        }
        FNTDprogress.updateProgress(1);
        
        this.updateTracks();
        
        if (IJ.getInstance() == null)
            System.out.println("\n - Tracks filtered: " + (nTracks - this.getNTracks()) + "\n");
    }
    
    

    /**
     * Removes tracks that are not present in given array of IDs
     * 
     * @param ft FNTDtracks object
     * @param trackID Array with track IDs
     * @return FNTDtracks object
     */
    public static FNTDtracks filterTrackID(FNTDtracks ft, double[] trackID) {
        ft.filterTrackID(trackID);

        return ft;
    }

    /**
     * Removes tracks that are not present in given array of IDs
     * 
     * @param trackID Array with track IDs
     */
    private void filterTrackID(double[] trackID) {
        int nChannels = getNChannels();
        int nFrames = getNFrames();
        int nSlices = getNSlices();
        
        for (int i = 0; i < nChannels; i++) {
            for (int j = 0; j < nFrames; j++) {
                for (int k = 0; k < nSlices; k++) {
                    for (FNTDParticle p : this.fntdParticles[i][j][k]) {
                        if (!ArrayUtils.contains(trackID, p.trackNo))
                        {
                            p.trackNo = 0;
                            p.prev_index = -1;
                            p.prev_slice = -1;
                            p.next_index = -1;
                            p.next_slice = -1;
                        } 
                    }
                }
            }
        }
    }
    
    
    /*************************************************************************
     * 
     * A C C E S S                                                           *
     * 
     *************************************************************************/
    /**
     * Checks whether the fntdParticles array is initialized
     * 
     * @return boolean
     */
    
    private boolean arrayInitilized() {
        return (!(fntdParticles == null));
    }

    /**
     * Checks whether the fntdParticles array is initialized at given slice
     * 
     * @param iSlice Slice number
     * @param jFrame Frame number
     * @param kChannel Channel number
     * @return boolean
     */
    private boolean memberInitilized(int iSlice, int jFrame, int kChannel) {
        if (withinBounds(iSlice, jFrame, kChannel))
            return (!(fntdParticles[kChannel][jFrame][iSlice] == null));
        
        return (false);
    }

    /**
     * Checks whether the given slice exists in the current image
     * 
     * @param iSlice Slice number
     * @param jFrame Frame number
     * @param kChannel Channel number
     * @return boolean
     */
    private boolean withinBounds(int iSlice, int jFrame, int kChannel) {
        int nSlices = getNSlices();
        int nFrames = getNFrames();
        int nChannels = getNChannels();
        return ((iSlice >= 0) && (iSlice < nSlices)
                && (jFrame >= 0) && (jFrame < nFrames)
                && (kChannel >= 0) && (kChannel < nChannels));
    }

    /**
     * Returns the FNTDParticle vector at the given slice
     * 
     * @param iSlice Slice number
     * @param jFrame Frame number
     * @param kChannel Channel number
     * @return Vector of FNTDParticle
     */
    public Vector<FNTDParticle> getTracks(int iSlice, int jFrame, int kChannel) {
        if (arrayInitilized() && memberInitilized(iSlice, jFrame, kChannel) && withinBounds(iSlice, jFrame, kChannel))
            return (fntdParticles[kChannel][jFrame][iSlice]);
        
        return (null);
    }

    /**
     * Returns the track numbers of the particles in the given slice
     * 
     * @param iSlice Slice number
     * @param jFrame Frame number
     * @param kChannel Channel number
     * @return Integer array
     */
    public int[] getTrackNo(int iSlice, int jFrame, int kChannel) {
        if (arrayInitilized() && memberInitilized(iSlice, jFrame, kChannel) && withinBounds(iSlice, jFrame, kChannel)) {
            int[] res = new int[fntdParticles[kChannel][jFrame][iSlice].size()];
            
            for (int i = 0; i < fntdParticles[kChannel][jFrame][iSlice].size(); i++)
                res[i] = fntdParticles[kChannel][jFrame][iSlice].elementAt(i).trackNo;
            
            return res;
        }
        return (null);
    }

    /**
     * Returns the x positions of the particles in the given slice
     * 
     * @param iSlice Slice number
     * @param jFrame Frame number
     * @param kChannel Channel number
     * @return Double array
     */
    public double[] getX(int iSlice, int jFrame, int kChannel) {
        if (arrayInitilized() && memberInitilized(iSlice, jFrame, kChannel) && withinBounds(iSlice, jFrame, kChannel)) {
            double[] res = new double[fntdParticles[kChannel][jFrame][iSlice].size()];
            
            for (int i = 0; i < fntdParticles[kChannel][jFrame][iSlice].size(); i++)
                res[i] = (double) fntdParticles[kChannel][jFrame][iSlice].elementAt(i).iX;
            
            return res;
        }
        return (null);
    }

    /**
     * Returns the y positions of the particles in the given slice
     * 
     * @param iSlice Slice number
     * @param jFrame Frame number
     * @param kChannel Channel number
     * @return Double array
     */
    public double[] getY(int iSlice, int jFrame, int kChannel) {
        if (arrayInitilized() && memberInitilized(iSlice, jFrame, kChannel) && withinBounds(iSlice, jFrame, kChannel)) {
            double[] res = new double[fntdParticles[kChannel][jFrame][iSlice].size()];
            
            for (int i = 0; i < fntdParticles[kChannel][jFrame][iSlice].size(); i++)
                res[i] = (double) fntdParticles[kChannel][jFrame][iSlice].elementAt(i).iY;
            
            return res;
        }
        return (null);
    }

    /**
     * Returns the velocity vectors in the x direction for the particles in the given slice
     * 
     * @param iSlice Slice number
     * @param jFrame Frame number
     * @param kChannel Channel number
     * @return Double array
     */
    public double[] getLx(int iSlice, int jFrame, int kChannel) {
        if (arrayInitilized() && memberInitilized(iSlice, jFrame, kChannel) && withinBounds(iSlice, jFrame, kChannel)) {
            double[] res = new double[fntdParticles[kChannel][jFrame][iSlice].size()];
            
            for (int i = 0; i < fntdParticles[kChannel][jFrame][iSlice].size(); i++)
                res[i] = (double) fntdParticles[kChannel][jFrame][iSlice].elementAt(i).lx;
            
            return res;
        }
        return (null);
    }

    /**
     * Returns the velocity vectors in the y direction for the particles in the given slice
     * 
     * @param iSlice Slice number
     * @param jFrame Frame number
     * @param kChannel Channel number
     * @return Double array
     */
    public double[] getLy(int iSlice, int jFrame, int kChannel) {
        if (arrayInitilized() && memberInitilized(iSlice, jFrame, kChannel) && withinBounds(iSlice, jFrame, kChannel)) {
            double[] res = new double[fntdParticles[kChannel][jFrame][iSlice].size()];
            
            for (int i = 0; i < fntdParticles[kChannel][jFrame][iSlice].size(); i++)
                res[i] = (double) fntdParticles[kChannel][jFrame][iSlice].elementAt(i).ly;
            
            return res;
        }
        return (null);
    }

    /**
     * Returns the zero-order intensity momenta for the particles in the given slice
     * 
     * @param iSlice Slice number
     * @param jFrame Frame number
     * @param kChannel Channel number
     * @return Double array
     */
    public double[] getM0(int iSlice, int jFrame, int kChannel) {
        if (arrayInitilized() && memberInitilized(iSlice, jFrame, kChannel) && withinBounds(iSlice, jFrame, kChannel)) {
            double[] res = new double[fntdParticles[kChannel][jFrame][iSlice].size()];
            
            for (int i = 0; i < fntdParticles[kChannel][jFrame][iSlice].size(); i++)
                res[i] = (double) fntdParticles[kChannel][jFrame][iSlice].elementAt(i).m0;
            
            return res;
        }
        return (null);
    }

    /**
     * Returns the first-order intensity momenta for the particles in the given slice
     * 
     * @param iSlice Slice number
     * @param jFrame Frame number
     * @param kChannel Channel number
     * @return Double array
     */
    public double[] getM1(int iSlice, int jFrame, int kChannel) {
        if (arrayInitilized() && memberInitilized(iSlice, jFrame, kChannel) && withinBounds(iSlice, jFrame, kChannel)) {
            double[] res = new double[fntdParticles[kChannel][jFrame][iSlice].size()];
            
            for (int i = 0; i < fntdParticles[kChannel][jFrame][iSlice].size(); i++)
                res[i] = (double) fntdParticles[kChannel][jFrame][iSlice].elementAt(i).m1;
            
            return res;
        }
        return (null);
    }

    /**
     * Returns the second-order intensity momenta for the particles in the given slice
     * 
     * @param iSlice Slice number
     * @param jFrame Frame number
     * @param kChannel Channel number
     * @return Double array
     */
    public double[] getM2(int iSlice, int jFrame, int kChannel) {
        if (arrayInitilized() && memberInitilized(iSlice, jFrame, kChannel) && withinBounds(iSlice, jFrame, kChannel)) {
            double[] res = new double[fntdParticles[kChannel][jFrame][iSlice].size()];
            
            for (int i = 0; i < fntdParticles[kChannel][jFrame][iSlice].size(); i++)
                res[i] = (double) fntdParticles[kChannel][jFrame][iSlice].elementAt(i).m2;
            
            return res;
        }
        return (null);
    }

    /**
     * Returns the third-order intensity momenta for the particles in the given slice
     * 
     * @param iSlice Slice number
     * @param jFrame Frame number
     * @param kChannel Channel number
     * @return Double array
     */
    public double[] getM3(int iSlice, int jFrame, int kChannel) {
        if (arrayInitilized() && memberInitilized(iSlice, jFrame, kChannel) && withinBounds(iSlice, jFrame, kChannel)) {
            double[] res = new double[fntdParticles[kChannel][jFrame][iSlice].size()];
            
            for (int i = 0; i < fntdParticles[kChannel][jFrame][iSlice].size(); i++)
                res[i] = (double) fntdParticles[kChannel][jFrame][iSlice].elementAt(i).m3;
            
            return res;
        }
        return (null);
    }

    /**
     * Returns the fourth-order intensity momenta for the particles in the given slice
     * 
     * @param iSlice Slice number
     * @param jFrame Frame number
     * @param kChannel Channel number
     * @return Double array
     */
    public double[] getM4(int iSlice, int jFrame, int kChannel) {
        if (arrayInitilized() && memberInitilized(iSlice, jFrame, kChannel) && withinBounds(iSlice, jFrame, kChannel)) {
            double[] res = new double[fntdParticles[kChannel][jFrame][iSlice].size()];
            
            for (int i = 0; i < fntdParticles[kChannel][jFrame][iSlice].size(); i++)
                res[i] = (double) fntdParticles[kChannel][jFrame][iSlice].elementAt(i).m4;
            
            return res;
        }
        return (null);
    }

    /**
     * Returns the particle discrimination nonParticleDiscriminationScore of the particles in the given slice
     * 
     * @param iSlice Slice number
     * @param jFrame Frame number
     * @param kChannel Channel number
     * @return Double array
     */
    public double[] getScore(int iSlice, int jFrame, int kChannel) {
        if (arrayInitilized() && memberInitilized(iSlice, jFrame, kChannel) && withinBounds(iSlice, jFrame, kChannel)) {
            double[] res = new double[fntdParticles[kChannel][jFrame][iSlice].size()];
            
            for (int i = 0; i < fntdParticles[kChannel][jFrame][iSlice].size(); i++)
                res[i] = (double) fntdParticles[kChannel][jFrame][iSlice].elementAt(i).nonParticleDiscriminationScore;
            
            return res;
        }
        return (null);
    }

    /**
     * Returns the total number of tracks
     * 
     * @return Integer
     */
    public int getNTracks() {
        return this.TrackNo;
    }

    /**
     * Returns the number of slices in the current image
     * 
     * @return Integer
     */
    public int getNSlices() {
        if (arrayInitilized()) return (fntdParticles[0][0].length);
        return 0;
    }

    /**
     * Returns the number of frames in the current image
     * 
     * @return Integer
     */
    public int getNFrames() {
        if (arrayInitilized()) return (fntdParticles[0].length);
        return 0;
    }

    /**
     * Returns the number of channels in the current image
     * 
     * @return Integer
     */
    public int getNChannels() {
        if (arrayInitilized()) return (fntdParticles.length);
        return 0;
    }

    /**
     * Returns the number of FNTDParticles in the given slice
     * 
     * @param iSlice Slice number
     * @param jFrame Frame number
     * @param kChannel Channel number
     * @return Integer
     */
    public int getNumberOfParticles(int iSlice, int jFrame, int kChannel) {
        if (arrayInitilized() && memberInitilized(iSlice, jFrame, kChannel) && withinBounds(iSlice, jFrame, kChannel))
            return (fntdParticles[kChannel][jFrame][iSlice].size());
        return 0;
    }

    /**
     * Returns the total number of FNTDParticles
     * 
     * @return Integer
     */
    public int getTotalNumberOfParticles() {
        if (arrayInitilized()) {
            int totalNumberOfTracks = 0;
            for (int i = 0; i < getNSlices(); i++) {
                for (int j = 0; j < getNFrames(); j++) {
                    for (int k = 0; k < getNChannels(); k++)
                        totalNumberOfTracks += fntdParticles[k][j][i].size();
                }
            }
            return (totalNumberOfTracks);
        }
        return 0;
    }

    /**
     * Checks of each particle whether the distance to its closest particle is
     * smaller than the critical distance.
     * 
     * @param criticalDistance Critical distance
     * @return Boolean array
     */
    public boolean[][][][] belowCriticalDistance(double criticalDistance) {

        boolean[][][][] belowCriticalDistance = new boolean[getNChannels()][getNFrames()][getNSlices()][];
        double[] x, y;
        for (int i = 0; i < getNChannels(); i++) {
            for (int j = 0; j < getNFrames(); j++) {
                for (int k = 0; k < getNSlices(); k++) {
                    int n = getNumberOfParticles(k, j, i);
                    if (n > 0) {
                        belowCriticalDistance[i][j][k] = new boolean[n];
                        Arrays.fill(belowCriticalDistance[i][j][k], false);
                        x = getX(k, j, i);
                        y = getY(k, j, i);
                        for (int l = 0; l < x.length - 1; l++) {
                            for (int m = l + 1; m < x.length; m++) {
                                if (Math.sqrt((x[l] - x[m]) * (x[l] - x[m]) + (y[l] - y[m]) * (y[l] - y[m])) <= criticalDistance) {
                                    belowCriticalDistance[i][j][k][l] = true;
                                    belowCriticalDistance[i][j][k][m] = true;
                                }
                            }
                        }
                    } else
                        belowCriticalDistance[i][j][k] = null;
                }
            }
        }
        return belowCriticalDistance;
    }
    
    /**
     * Converts the given FNTDtracks object to an array containing 
     * the information of the FNTDParticles
     * @param ft FNTDtracks object
     * @return Double array. CAREFUL: Indices for frames, slices, channels
     * change from zero-based (Java style) to one-based (R or similar style)
     */
    public static double[][] toArray(FNTDtracks ft) {
        return ft.toArray();
    }
    
    /**
     * Converts the given FNTDtracks object to an array containing 
     * the information of the FNTDParticles
     * @return Double array. CAREFUL: Indices for frames, slices, channels
     * change from zero-based (Java style) to one-based (R or similar style)
     */
    public double[][] toArray() {
        int nSlices = getNSlices();
        int nFrames = getNFrames();
        int nChannels = getNChannels();
        
        double[][] array = new double[9][this.getTotalNumberOfParticles()];

        int index = 0;
        for (int i = 0; i < nChannels; i++) {
            for (int j = 0; j < nFrames; j++) {
                for (int k = 0; k < nSlices; k++) {
                    for (FNTDParticle p : this.fntdParticles[i][j][k]) {
                        array[0][index] = k + 1;
                        array[1][index] = j + 1;
                        array[2][index] = i + 1;
                        array[3][index] = p.trackNo;
                        array[4][index] = p.iX;
                        array[5][index] = p.iY;
                        array[6][index] = p.lx;
                        array[7][index] = p.ly;
                        array[8][index] = p.m0;
                        index++;
                    }
                }
            }
        }
        return array;
    }

    /**
     * Converts the given FNTDtracks object to an RoiManager containing ROIs
     * around the FNTDParticle centra.
     * @param radius Radius around the maximum pixel value
     * @return RoiManager array
     */
    private Vector[][][] toROIs(double radius) {
        if (!arrayInitilized()) return (null);

        int nSlices = getNSlices();
        int nFrames = getNFrames();
        int nChannels = getNChannels();

        Vector[][][] rois = new Vector[nChannels][nFrames][nSlices];
        double[] x, y;
        for (int i = 0; i < nChannels; i++) {
            for (int j = 0; j < nFrames; j++) {
                for (int k = 0; k < nSlices; k++) {
                    rois[i][j][k] = new Vector<>();
                    x = getX(k, j, i);
                    y = getY(k, j, i);
                    if (!(x == null)) {
                        for (int l = 0; l < x.length; l++) {
                            if (x[l] != 0.0 && y[l] != 0.0)
                                rois[i][j][k].add(new OvalRoi(x[l] - radius / 2, y[l] - radius / 2, radius, radius));
                            else
                                rois[i][j][k].add(new OvalRoi(0, 0, 0, 0));
                        }
                    }
                }
            }
        }
        return rois;
    }

    /**
     * Converts the given stack of slices to a Vector of Trajectories
     * 
     * @param jFrame Frame number
     * @param kChannel Channel number
     * @return Vector of Trajectory
     */
    public Vector<Trajectory> toTrajectories(int kChannel, int jFrame) {

        Vector<Trajectory> trajectories = new Vector<>();
        Vector<Integer> trackNos = new Vector<>();

        for (int i = 0; i < getNSlices(); i++) {
            for (FNTDParticle p : this.fntdParticles[kChannel][jFrame][i]) {
                if (p.trackNo > 0) {
                    if (!trackNos.contains(p.trackNo)){
                        trackNos.add(p.trackNo);
                        trajectories.add(new Trajectory(p.trackNo));
                    }
                    
                    trajectories.elementAt(trackNos.indexOf(p.trackNo)).addParticle(p);
                }
            }
        }
        return trajectories;
    }

    /**
     * Converts the given stack ImagePlus to a mask
     * 
     * @param imp ImagePlus object
     * @param ft FNTDTracks object
     * @param radius Radius around the maximum pixel value
     * @param criticalDistance Critical distance
     * @return ImagePlus object
     */
    public static ImagePlus toMask(ImagePlus imp,
            FNTDtracks ft,
            double radius,
            double criticalDistance) {

        ImagePlus tmpImp = ft.toMask(imp, radius, criticalDistance);
        Duplicator dp = new Duplicator();
        ImagePlus resImp = dp.run(tmpImp);

        return resImp;
    }

    /**
     * Converts the given stack ImagePlus to a mask
     * 
     * @param imp ImagePlus object
     * @param ft FNTDTracks object
     * @param radius Radius around the maximum pixel value
     * @param criticalDistance Critical distance
     * @return ImagePlus object
     */
    private ImagePlus toMask(ImagePlus imp,
            double radius,
            double criticalDistance) {

        // Get ROIs from tracks and
        Vector[][][] rois = toROIs(radius);
        boolean[][][][] belowCriticalDistance = belowCriticalDistance(criticalDistance);

        ImageProcessor ip;

        ImagePlus resImp = NewImage.createByteImage("Mask",
                imp.getWidth(),
                imp.getHeight(),
                imp.getImageStackSize(),
                NewImage.FILL_BLACK);

        for (int i = 0; i < rois.length; i++) {
            for (int j = 0; j < rois[0].length; j++) {
                for (int k = 0; k < rois[0][0].length; k++) {

                    if (!rois[i][j][k].isEmpty()) {

                        resImp.setSlice(k + 1);
                        resImp.setT(j + 1);
                        resImp.setC(i + 1);

                        ip = resImp.getProcessor();
                        ip.setValue(255);
                        for (int l = 0; l < rois[i][j][k].size(); l++) {
                            if (belowCriticalDistance[i][j][k][l] == false) {
                                Roi curRoi = (Roi)rois[i][j][k].get(l);
                                OvalRoi or = new OvalRoi(curRoi.getBounds().x,
                                        curRoi.getBounds().y,
                                        curRoi.getBounds().width,
                                        curRoi.getBounds().height);
                                ip.setRoi(or);
                                ip.fill(or);
                            }
                        }

                        resImp.setProcessor(ip);
                    }

                    System.out.println("Did slice " + Integer.toString(k + 1) + ", frame " + Integer.toString(j + 1) + ", channel " + Integer.toString(i + 1) + ".");
                }
            }
        }
        return resImp;
    }

    /**
     * Class to store trajectories and their respective properties
     */
    public static class Trajectory {

        public int start_slice = Integer.MAX_VALUE;
        public int stop_slice  = -1;
        public Vector<FNTDParticle> existing_particles;         // holds all particles of this trajetory in order
        public double lx = 0, ly = 0, m0 = 0, m2 = 0;           // Hold track average values for relinking of tracks
        int length = 0;                                         // number of frames this trajectory spans on

        int serial_number;                                      // serial number of this trajectory (for report and display)

        public Trajectory(Vector<FNTDParticle> particles) {
            this.existing_particles = particles;
            // the length is the last trajectory frame - the first frame (first frame can be 0) 
            this.length = this.existing_particles.lastElement().slice
                    - this.existing_particles.firstElement().slice;
            
            for (FNTDParticle p : particles) {
                this.m0 += p.m0;
                this.m2 += p.m2;
                this.lx += p.ly;
                this.ly += p.lx;
            }
            
            if (particles.size() > 1) {
                this.m0 /= particles.size();
                this.m2 /= particles.size();
                this.lx /= (particles.size() - 1);
                this.ly /= (particles.size() - 1);
            }
        }

        public Trajectory(int serial_number) {
            this.existing_particles = new Vector<>();
            this.length = 0;
            this.serial_number = serial_number;
        }
        
        public void addParticle(FNTDParticle p) {
            this.existing_particles.add(p);
            
            this.m0 = (this.m0 * (this.existing_particles.size() - 1) + p.m0) / this.existing_particles.size();
            this.m2 = (this.m2 * (this.existing_particles.size() - 1) + p.m2) / this.existing_particles.size();
            
            if (this.existing_particles.size() == 1) {
                this.lx = 0;
                this.ly = 0;
            } else {
                this.lx = (this.lx * (this.existing_particles.size() - 1) + p.lx) / (this.existing_particles.size() - 1);
                this.ly = (this.ly * (this.existing_particles.size() - 1) + p.ly) / (this.existing_particles.size() - 1);
            }

            if (p.slice > this.stop_slice)
                this.stop_slice = p.slice;
            if (p.slice < this.start_slice)
                this.start_slice = p.slice;
                
            this.length = this.stop_slice - this.start_slice + 1;
        }
    }
}
