FNTD.get.track.fits <- function(x, voxel.size, surface.param, optical.param, image.size, depth.refinement, amplitude.endpoint.mod, detour.range) {
  
  if(class(x) != "data.frame" & class(x) != "FNTDjImage")
    stop("Object x has to be of class \"FNTDjImage\" or \"data.frame\".")
  
  if(class(x) == "FNTDjImage") {
    df  <- FNTD.get.tracks(x)
    n   <- dim(x)[1:3]
    n.x <- n[1]
    n.y <- n[2]
    n.z <- n[3]
  } else {
    df <- x
    if (missing(image.size))
      stop("Please give image size (n.x, n.y) and number of slices (n.slices) when using trackspot dataframe.")
    
    n.x <- image.size$n.x
    n.y <- image.size$n.y
    n.z <- image.size$n.slices
  }
  
  exp.fit <- function(b, a, x, y = NULL, error = TRUE) {
    rhat <- a + (1-a)*(1 - exp(-b*x))
    
    if (error)
      return(sum((y - rhat)^2))
    else
      return(rhat)
  }
  
  depth.ref.range <- 2.5
  
  # remove trackspots that do not belong to any track
  df <- df[df$track.no != 0,]
  
  cat("Compute linear regression on trajectories...\n")
  
  # initialise return data frame
  fit.results <- data.frame()
  
  # scale to account for z-pixel size and correct for spherical aberrations
  df$zpos     <- FNTD.find.depth(df, fit.param  = surface.param,
                                 voxel.size     = voxel.size,
                                 optical.param  = optical.param)
  
  cv          <- paste(df$track.no, df$frame.no)
  cv.unique   <- unique(cv)
  c           <- 0
  d           <- 0
  e           <- 0
  f           <- 0
  for (i in 1:length(cv.unique)) {
    # i <- 1
    ii      <- cv.unique[i] == cv
    
    if(sum(ii) >= 2) {
      
      if (i%%100==0)
        cat("...doing track ", i, " of ", length(cv.unique), "...\n")
      
      
      # GET SUBSET FOR ONE TRACK
      df.sub   <- subset(df, ii)
      
      ## <Projected and CSDA length estimation>
      df.sub      <- FNTD.get.proc.track.length(df.sub, voxel.size = voxel.size)
      p.last      <- tail(df.sub,1)
      
      df.ref      <- tail(df.sub,depth.ref.range/voxel.size$z.um)
      amp.mean    <- mean(df.ref$amplitude, na.rm=TRUE)
      amp.sd      <- sd  (df.ref$amplitude, na.rm=TRUE)
      
      x.pos.adj   <- p.last$x
      y.pos.adj   <- p.last$y
      z.pos.adj   <- p.last$zpos
      csda.length <- p.last$length
      
      if (depth.refinement != "None" & !is.na(p.last$amplitude)) {
        if (p.last$amplitude < amp.mean*amplitude.endpoint.mod) { # Find endpoint within trackpoints => SHOULD ALWAYS BE THE CASE
          p.lat           <- tail(df.ref[df.ref$amplitude >= amplitude.endpoint.mod*amp.mean,], 1)
          
          do.full  = depth.refinement == "Full"
          do.SD    = depth.refinement == "SD"
          do.Slope = depth.refinement == "Slope"
          
          if (do.full) {
            a               <- p.last$amplitude/amp.mean
            fit             <- optim(.5, exp.fit, x = max(df.ref$zpos) - df.ref$zpos, y = df.ref$amplitude/amp.mean, a = a,
                                     method="BFGS", control=list(reltol=1e-3))  # A = a + (1-a)*(1-exp(-b*z))
            z.pos.adj       <- max(df.ref$zpos) + log(1 - (amplitude.endpoint.mod - a)/(1 - a))/fit$par[1]
            
            if (is.na(z.pos.adj) | is.nan(z.pos.adj) | abs(z.pos.adj - p.last$zpos) > depth.ref.range)
              do.SD = T
          }
          if (do.SD) {
            p.lat.sd        <- tail(df.ref[df.ref$amplitude >= amp.mean - amp.sd,], 1)
            if (length(p.lat.sd$zpos) > 0)
              z.pos.adj     <- p.lat.sd$zpos
            else
              f <- f + 1
          }
          if (do.Slope) {
            amplitude.slope <- (p.last$zpos - p.lat$zpos)/(p.last$amplitude - p.lat$amplitude) # dz/dA
            z.pos.adj       <- (amp.mean*amplitude.endpoint.mod - p.lat$amplitude) *
              amplitude.slope + p.lat$zpos #z_m = z1 + dz/dA * (Am - A1)
          }
          
          d <- d + 1
        
          if (is.na(z.pos.adj) | is.nan(z.pos.adj) | abs(z.pos.adj - p.last$zpos) > depth.ref.range) { # When z.pos.adj failed
            x.pos.adj   <- p.lat$x
            y.pos.adj   <- p.lat$y
            z.pos.adj   <- p.lat$zpos
            csda.length <- p.lat$length
            
            e <- e + 1
          } else {
            x.pos.adj   <- approx(x = df.sub$z, y = df.sub$x,      z.pos.adj)$y
            y.pos.adj   <- approx(x = df.sub$z, y = df.sub$y,      z.pos.adj)$y
            csda.length <- approx(x = df.sub$z, y = df.sub$length, z.pos.adj)$y
          }
        }
        c <- c + 1
      }
      
      fit.x       <- summary(fastLm(x * voxel.size$x.um ~ zpos, df.sub))
      fit.y       <- summary(fastLm(y * voxel.size$y.um ~ zpos, df.sub))
      proj.length <- z.pos.adj*sqrt(1 + fit.x$coefficients[2,1]^2 + fit.y$coefficients[2,1]^2)
      
      ## <Calculate Detour factor>
      if (length(df.sub[df.sub$zpos <= z.pos.adj,]$zpos) > detour.range + 3) {
        fit.x.fdt <- summary(fastLm(x * voxel.size$x.um ~ zpos, df.sub[1:detour.range,]))
        fit.y.fdt <- summary(fastLm(y * voxel.size$y.um ~ zpos, df.sub[1:detour.range,]))
        
        # Projected last position onto vector of first detour.range trackspots
        Q <- c(x.pos.adj * voxel.size$x.um - fit.x.fdt$coefficients[1,1],
               y.pos.adj * voxel.size$y.um - fit.y.fdt$coefficients[1,1],
               z.pos.adj) # Vector to last (extended) particle
        V <- c(fit.x.fdt$coefficients[2,1], fit.y.fdt$coefficients[2,1], 1)
        
        # Vector projection
        detourp <- sqrt(sum((sum(Q*V)/sum(V*V)*V)^2))/csda.length
        detouro <- sqrt(sum((Q - sum(Q*V)/sum(V*V)*V)^2))/csda.length
      } else {
        detourp <- NA
        detouro <- NA
      }
      
      ## <--->
      
      if (surface.param$theta.cor) {
        corr <- FNTD.surface.theta.correction(fit.x$coefficients[2,1], fit.y$coefficients[2,1], surface.param)
        fit.x$coefficients[2,1] <- corr[[1]]
        fit.y$coefficients[2,1] <- corr[[2]]
      }
      
      fit.results <- rbind(fit.results,
                           data.frame(track.no     = unique(df.sub$track.no),
                                      frame.no     = unique(df.sub$frame.no),
                                      csda.length  = csda.length,
                                      proj.length  = proj.length,
                                      detour       = list('p' = detourp, 'o' = detouro),
                                      xz.slope     = fit.x$coefficients[2,1],
                                      xz.d.slope   = fit.x$coefficients[2,2],
                                      yz.slope     = fit.y$coefficients[2,1],
                                      yz.d.slope   = fit.y$coefficients[2,2],
                                      xz.int.um    = fit.x$coefficients[1,1],
                                      xz.d.int.um  = fit.x$coefficients[1,2],
                                      yz.int.um    = fit.y$coefficients[1,1],
                                      yz.d.int.um  = fit.y$coefficients[1,2],
                                      length.vec   = tapply(df.sub$length,  df.sub$track.no,  as.list)))
    }
  }
  
  # Calculate the angles from the fitted slopes
  fit.results <- FNTD.get.track.angles(fit.results)
  
  if (depth.refinement != "None")
    cat("Tracks adjusted: ", c, ". Within track: ", d, ". Outside track: ", c-d, "\n", "Range refinement failed: (", e, ", ", f, ")\n", sep="") 
  cat("\n")
  
  return(fit.results)
}