#' generate.simple.tracks
#' 
#' Returns a grid of tracks, perpendicular and of constant
#' intensity, same for all frames, all track spots circular
#' with constant intensity
#'
#' @param x          x positions for tracks
#' @param y          y positions for tracks
#' @param n.slices   number of slices
#' @param n.frames   number of frames
#' @param intensity  track intensity
#'
#' @return data-frame for use with FNTD.create.image
generate.simple.tracks <- function(x, y, n.slices, n.frames, intensity, radius, add.to.background){
  df.tracks <- expand.grid(x      = x,
                           y      = y,
                           z      = 1:n.slices,
                           t      = 1:n.frames,
                           type   = "circle",
                           param1 = intensity,
                           param2 = radius,
                           param3 = add.to.background, 
                           param4 = NA,
                           param5 = NA)
  return(df.tracks)
}

#' i2v converts ImageJ indices to linear index
#'
#' @param i.x      x index of voxel (one-based)
#' @param i.y      y index of voxel (one-based)
#' @param i.slice  slice index of voxel (one-based)
#' @param i.frame  frame index of voxel (one-based)
#' @param size     list with image dimensions (n.x, n.y, n.slices, n.frames)
#'
#' @return index in vector
i2v <- function(i.x, i.y, i.slice, i.frame, size){
   return((i.frame - 1) * size$n.slices * size$n.y * size$n.x +
          (i.slice - 1) * size$n.y * size$n.x +
          (i.y - 1) * size$n.x +
           i.x)
}

#' add circle
#'
#' @param dat     intensity vector
#' @param track   respective row of data frame
#' @param size    list with image dimensions (n.x, n.y, n.slices, n.frames)
#'
#' @return updated intensity vector
add.circle <- function(dat, track, size){
  radius  <- track$param2
  grid    <- expand.grid(x = -radius:radius,
                         y = -radius:radius)
  ii      <- grid$x^2+grid$y^2 <= radius
  grid    <- grid[ii,]
  idx  <- i2v(grid$x + track$x, 
              grid$y + track$y, 
              track$z,
              track$t,
              size)
  if(track$param3){
    dat[idx] <- dat[idx] + track$param1
  }else{
    dat[idx] <- track$param1
  }
  return(dat)
}

#' FNTD.create.image returns a FNTDjimage object with
#' given properties
#'
#' @param size       list with image dimensions (n.x, n.y, n.slices, n.frames)
#' @param background list with background propertied (mean, sd [<0: Poisson, >0: Gauss])
#' @param tracks     data frame with track entries (x, y, z, t index, type, 5 parameters)
#'
#' @return FNTDjimage object
FNTD.create.image <- function(size, background, df.tracks){
  
  n.data <- size$n.x * size$n.y * size$n.slices * size$n.frames
  
  # Intensity vector
  dat   <- numeric(n.data)
  
  if( background$sd == 0){ # No noise
    dat    <- rep(background$mean, n.data)
  }
  if( background$sd < 0){  # Poisson noise
    dat    <- rpois(n = n.data, 
                       lambda = background$mean)
  }
  if( background$sd > 0){  # Custom Gaussian noise
    dat    <- rnorm(n = n.data, 
                       mean = background$mean, 
                       sd = background$sd)
  }
  
  # Add tracks
  for(i in 1:nrow(df.tracks)){
     dat <- add.circle(dat   = dat,
                       track = df.tracks[i,],
                       size  = size)  
  }

  jimg <- FNTDjImage(data = dat, 
                     dim  = c(size$n.x, size$n.y, size$n.slices, size$n.frames, 1), 
                     title = "New test image")
  
  return(jimg)
}