\name{FNTD.get.track.fits}
\alias{FNTD.get.track.fits}

\title{Apply linear regression analysis on particle tracks}

\description{Returns a data frame containing the fit results as well as the derived polar and azimuthal angle for all trajectories.}

\usage{FNTD.get.track.fits(x, voxel.size    = list(x.um = .1,
                                                   y.um = .1,
                                                   z.um = .1),
                              surface.param = list(z.offset = 0,
                                                   x.slope = 0,
                                                   y.slope = 0,
                                                   image.offset = 0,
                                                   theta.cor = FALSE),
                              optical.param = list(n1 = 1.517,
                                                   n2 = 1.7659,
                                                   NumAp = 1.40),
                              image.size    = list(n.x = 1024,
                                                   n.y = 1024,
                                                   n.slices = 30),
                              depth.refinement = FALSE)
}

\arguments{
  \item{x}{Either an FNTDjImage object or a dataframe containing trackspot properties (as returned by \code{\link{FNTD.get.trackspot.properties}}).}
  \item{voxel.size}{The voxel dimensions without ASF corrections given in um.}
  \item{surface.param}{Fit parameters obtained from a linear fit of the surface reflection. Used to correct tilting during FNTD scanning. Also corrects the theta angles of the tracks when \code{theta.cor} is TRUE.}
  \item{optical.param}{Refractive indices and numerical aperture for the scanning setup. Used for correction of aberrations due to the refractive index mismatch between the FNTD and immersion medium.}
  \item{image.size}{Image dimensions in pixels. Only required when x is a dataframe.}
  \item{depth.refinement}{Deploy a routine that aims to refine the endpoint of each trajectory by weighing the intensities of the last trackpoints.}
}

\details{
Note that this routine is called from within FNTD.get.track.properties.

If \code{x} is an image containing track data, linear regression analysis between the track spot position in x resp. y and the z-position is performed (using \code{\link{lm}}). The relative z-positions are given by the distance between the slices multiplied with the scaling factor \code{pixel.ratio.zx} and the parameter \code{ASF}. The default value \code{ASF=1.2} accounts for the error in the position of the focal point due to spherical aberration. This can be used when dealing with positive transitions in refractive index (e.g. from oil into FNTD). For negative (e.g. oil-water-FNTD) or more complexs situations, see M. Niklas et al. (2013).\cr 

In comparison to the linear fit x ~ z, the second fit y ~ z is scaled with \code{pixel.ratio.yx} to take into account possible different pixel size in x- and y-direction.}

\value{
 \code{FNTD.get.track.fits} returns a data frame containing the trajectory slopes, intercepts, the polar and azimuthal angles and their respective standard uncertainities (1.s.d) for each track plus its entrance and exits point coordinates and plain.\\
 
 IMPORTANT: The values given in the returned data frame are given with respect to the actual pixel size and ratios (i.e. in um), while information from \code{\link{FNTD.get.tracks}} etc. are given in pixel / slice indices!\\
 
  The fit parameters of the slopes are used to compute the angular orientation of the track. The uncertainties of the polar angle \code{theta} (measured to z-axis) and azimuthal angle \code{phi} (measured to x-axis) is calculated according to error propagation. 
}

\author{
M. Heinz, Steffen Greilich, Jasper J. M. Kouwenberg
}

\references{
Van Elburg H. J. et al. (2007), Improved correction of axial geometrical distortion in index mismatched fluorescent confocal microscopic images using high-aperture objective lenses. Journal of Microscopy, Vol. 228(1), 45-54.

Niklas, M. et al. (2013), Ion track reconstruction in 3D using alumina-based fluorescent nuclear track detectors. Physics in Medicine and Biology, 2013, 58, N251-N266.

Klimpki, G. (2014),  Fluence-based dosimetry using fluorescence nuclear track detectors, M.Sc. thesis, Department of Physics and Astronomy, Heidelberg University. URL: http://www.ub.uni-heidelberg.de/archiv/17356.
}

\examples{
\dontrun{
jimg <- FNTD.read.image(system.file("extdata", "mh2003.tif.zip", package = "FNTD"))

# Import and assign track data (from external tracker/linker)
jimg <- FNTD.import.tracks(jimg, 
                           system.file("extdata", "mh2003_trajectories.csv", package = "FNTD"),
                           format = "MosaicLinker")
# Use size/ASF for FXR700RG
fit.results <- FNTD.get.track.fits( jimg, 
                                    voxel.size.x.um = 100/504,
                                    voxel.ratio.yx  = 1, 
                                    voxel.ratio.zx  = 1700/320, #500/131.79, 
                                    ASF             = 1.2)
}
}