\name{read.FNTD.images}
\alias{read.text.image}
\alias{read.lsm.image}
\alias{read.raw.image}
\alias{read.tif.image}
\alias{get.FNTD.image}
\alias{get.substack}

\title{
Read FNTD images
}

\description{
Reads image data for FNTDs. Several input formats are possible, such as text as exported by ImageJ (most likely from Zeiss LSM710 data converted), TIFF, and LSM. If given, the positions of tracks detected by the ImageJ particle tracker (MOSAIC plugin), are read and added. For large images, a pre-read crop can be applied to import a part of the image only.\cr For very large images, read.raw.image should be used instead. In this case, the image data in the file must be provided as stream of 16bit integer raw data (use ImageJ export function). Data are loaded as compact as possible into the memory and only a pointer will be handed over to R. It is not possible to plot or apply functions on these big pictures but sections can very fast cut out as FNTD images using \code{get.FNTD.image}.\cr
Images stacks can also be read in LSM format. Besides plotting and some other function, most routines are however not yet equipped for handling image stacks. Therefore use \code{get.substack} to extract single (or multiple) frames.
}

\usage{
read.lsm.image(filename, first.frame = NULL, last.frame = NULL)
read.tif.image(filename, first.frame = NULL, last.frame = NULL)
read.text.image(filename, filename.detected.tracks = NULL, x.min = NULL, x.max = NULL, y.min = NULL, y.max = NULL, rescans = 0, dwell.time.us = 0, laser.power.percent = 0)
read.raw.image(filename, x.size, y.size, type = "16bit")
get.FNTD.image(bimg, x.min, y.min, x.max, y.max, unprotected = FALSE)
get.substack(img, z.start, z.end)
}

\arguments{
  \item{filename}{ filename of text image data}
  \item{filename.detected.tracks}{ filename of detected tracks data}
  \item{first.frame}{ first frame to be read}
  \item{last.frame}{ last frame to be read}
  \item{x.min}{ left x corner for importing only part of an image}
  \item{x.max}{ right x corner for importing only part of an image}
  \item{y.min}{ lower y corner for importing only part of an image}
  \item{y.max}{ upper y corner for importing only part of an image}
  \item{rescans}{ Number of rescans during read-out, needed for frequency conversion}
  \item{dwell.time.us}{ Dwell-time (in us) during read-out, needed for frequency conversion}
  \item{laser.power.percent}{ Laser power (in percent) during read-out, needed for frequency conversion}
  \item{x.size}{ size in x for raw image}
  \item{y.size}{ size in y for raw image}
  \item{type}{ type of raw data, only 16bit integer yet.}
  \item{bimg}{object of class \code{FNTDbigimage}}
  \item{unprotected}{if TRUE, image with more than 1e6 pixels can be cut out as FNTD images. This is a protection as unintented cutting of very big section can stall the system}
  \item{img}{object of class \code{FNTDimage}}
  \item{z.start}{first z slice to extract}
  \item{z.end}{last z slice to extract}
}

\details{
The routines obey the zero-based image coordinates from ImageJ, the z index however starts at 1. For importing a part of an image all four arguments (x.min, x.max, y.min, y.max) have to be provided. In case of TIFF or text format, no metadata (dwelltime, laser power etc.) will be available!}

\value{
An \code{FNTDimage} object, an \code{FNTDbigImage} object in case of \code{read.raw.image}.
}


\author{
Steffen Greilich
}

\examples{
\dontrun{
filename <- system.file("extdata", "sg35044_z3.lsm", package = "FNTD")
img      <- read.lsm.image(filename)                              # Read image stack
plot(img, tracks)                                                 # Plot image stack
plot(get.substack(img, 1, 1))                                     # Plot first slice only

imgs   <- lapply( 1:img@sizeZ,                                    # Reorganize stack as list of single images
                  function(x, img){get.substack(img, x, x)}, 
                  img)                                        

plot(imgs[[1]])                                                   # Plot first image slice

avg.counts <- unlist( lapply( imgs, mean))                        # Get mean counts for all slices

bkg.imgs   <- lapply(imgs, get.background, length.parameter = 50) # Get background and 
bkg.avg    <- unlist( lapply( bkg.imgs, mean))                    # mean background intensity for all slices
}
}
